from django.contrib.auth.models import User
from django.urls import reverse

import hashlib
import uuid

from scidatacontainer_db.models import DataSet
from . import TESTDIR

from rest_framework.test import APITestCase

VIEW_NAME = "scidatacontainer_db:api:dataset-list"


class ApiDataSetListTest(APITestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = User.objects.create_user("testuser")

    def _post(self, url, **kwargs):
        self.client.force_authenticate(self.user)
        response = self.client.post(url, **kwargs)
        return response

    def _get(self, url, **kwargs):
        self.client.force_authenticate(self.user)
        response = self.client.get(url, **kwargs)
        return response

    def _create_test_dataset(self):
        filename = TESTDIR + "example.zdc"
        response = self._post(reverse(VIEW_NAME),
                              data={"uploadfile":
                                    open(filename, "rb")
                                    }
                              )
        self.assertEqual(response.status_code, 201)
        self.id = DataSet.objects.all()[0].id
        self.hash = hashlib.sha256(open(filename, "rb").read()).hexdigest()

    def test_view_url_exists_at_desired_location(self):
        response = self._get("/api/datasets/")
        self.assertEqual(response.status_code, 200)

    def test_view_url_accessible_by_name(self):
        response = self._get(reverse(VIEW_NAME))
        self.assertEqual(response.status_code, 200)

    def test_http_method(self):
        response = self._get(reverse(VIEW_NAME))
        self.assertEqual(response.status_code, 200)

    def test_dataset_list(self):
        response = self._get(reverse(VIEW_NAME))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data), 0)

        self._create_test_dataset()
        response = self._get(reverse(VIEW_NAME))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data), 1)

    def test_dataset_filter(self):
        self.test_dataset_list()

        response = self._get(reverse(VIEW_NAME),
                             data={"id": self.id})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data), 1)

        response = self._get(reverse(VIEW_NAME),
                             data={"id": uuid.uuid4()})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data), 0)

        response = self._get(reverse(VIEW_NAME),
                             data={"title__contains": "image"})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data), 1)

        response = self._get(reverse(VIEW_NAME),
                             data={"title__contains": "42"})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data), 0)



from django.contrib.auth.models import User
from django.test import TestCase
from django.urls import reverse

from knox.models import AuthToken


class KeysTest(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = User.objects.create_user("testuser")

    def _post(self, url, **kwargs):
        self.client.force_login(self.user)
        response = self.client.post(url, **kwargs)
        return response

    def _get(self, url, **kwargs):
        self.client.force_login(self.user)
        response = self.client.get(url, **kwargs)
        return response

    def test_view_url_exists_at_desired_location(self):
        response = self._get("/keys/")
        self.assertEqual(response.status_code, 200)

    def test_view_url_accessible_by_name(self):
        response = self._get(reverse("scidatacontainer_db:ui-keys"))
        self.assertEqual(response.status_code, 200)

    def test_template_keys(self):
        response = self._get(reverse("scidatacontainer_db:ui-keys"))
        self.assertTemplateUsed(response, "scidatacontainer_db/apikeys.html")

    def test_create_keys(self):
        response = self._get(reverse("scidatacontainer_db:ui-keys"))
        self.assertContains(response,
                            "class=\"btn-close\"",
                            count=0,
                            status_code=200)

        before = len(AuthToken.objects.all())
        response = self._post(reverse("scidatacontainer_db:ui-keys"),
                              data={"create": True})
        self.assertEqual(response.status_code, 302)
        after = len(AuthToken.objects.all())
        self.assertEqual(before + 1, after)

    def test_delete_keys(self):
        self.test_create_keys()
        before = len(AuthToken.objects.all())
        digest = (AuthToken.objects.filter(user=self.user)[0]).digest

        response = self._post(reverse("scidatacontainer_db:ui-keys"),
                              data={"delete": digest})

        self.assertEqual(response.status_code, 302)
        after = len(AuthToken.objects.all())
        self.assertEqual(before - 1, after)
        self.assertEqual(0, len(AuthToken.objects.filter(digest=digest)))



from django.test import TestCase
from scidatacontainer_db.models import ContainerType


class ContainerTypeTest(TestCase):

    def test_containerType_creation(self):
        c = ContainerType(name="TestType")
        c.save()
        self.assertTrue(isinstance(c, ContainerType))
        self.assertEqual(c.name, "TestType")
        self.assertEqual(str(c), "TestType")

        c = ContainerType(name="TestType2",
                          id="testID",
                          version="0.1")
        c.save()
        self.assertTrue(isinstance(c, ContainerType))
        self.assertEqual(c.name, "TestType2")
        self.assertEqual(c.id, "testID")
        self.assertEqual(c.version, "0.1")
        self.assertEqual(str(c), "TestType2, v0.1")

    def test_containerType_creation_from_dict(self):
        c = ContainerType.to_ContainerType("TestType3")
        c.save()
        self.assertTrue(isinstance(c, ContainerType))
        self.assertEqual(c.name, "TestType3")

        c = ContainerType.to_ContainerType({"name": "TestType4",
                                            "id": "testID2",
                                            "version": "0.2"})
        c.save()
        self.assertTrue(isinstance(c, ContainerType))
        self.assertEqual(c.name, "TestType4")
        self.assertEqual(c.id, "testID2")
        self.assertEqual(c.version, "0.2")

        rejected = False
        try:
            c = ContainerType.to_ContainerType([{"id": "testID2",
                                                "version": "0.2"}])
        except AssertionError as e:
            rejected = True
            self.assertEqual(str(e),
                             "ContainerType needs to be a " +
                             "string (name) or a dictionary.")
        self.assertTrue(rejected)

        rejected = False
        try:
            c = ContainerType.to_ContainerType({"id": "testID2",
                                                "version": "0.2"})
        except AssertionError as e:
            rejected = True
            self.assertEqual(str(e),
                             "ContainerType requires name attribute.")
        self.assertTrue(rejected)

        rejected = False
        try:
            c = ContainerType.to_ContainerType({"name": "TestType4",
                                                "id": "testID2"})
        except AssertionError as e:
            rejected = True
            self.assertEqual(str(e),
                             "ContainerType requires version" +
                             "attribute if id is given")
        self.assertTrue(rejected)



from django.core.exceptions import PermissionDenied
from django.http import FileResponse

import io

from scidatacontainer.tests import get_test_container

from .utils import APIResponse, MetaDBError
from .models import DataSet


def parse_test_data(uuid: str):
    """
    Check the last characters of a test UUID and raise a corresponding
    exception.

    :param uuid: UUID as str

    :raises scidatacontainer_db.MetaDBError: The exception matching the error
    code of the last 3 characters of the UUID.
    """
    if uuid.endswith("409"):
        raise MetaDBError({"error_code": 409,
                           "msg": "Dataset is marked complete. " +
                                  "No further changes allowed."})
    if uuid.endswith("403"):
        raise MetaDBError(
            {"error_code": 403,
             "msg": "You don't have permission to update this dataset."})

    if uuid.endswith("301"):
        pass


def get_test_data(uuid: str):
    dc = get_test_container()

    if uuid.endswith("301"):
        dc.freeze()
        DataSet(uuid=uuid.uuid4())
    return


def download_test_dataset(uuid: str):
    if uuid.endswith("204"):
        return APIResponse("", status=204, reason="DataSet was deleted")

    if uuid.endswith("404"):
        return APIResponse("", status=404, reason="No DataSet with UUID=" +
                                                  uuid + "found!")

    if uuid.endswith("403"):
        raise PermissionDenied

    container = get_test_container()
    if uuid.endswith("301"):
        container["content.json"]["replaces"] = uuid
        r = FileResponse(io.BytesIO(container.encode()))
        r.status_code = 301
        return r

    r = FileResponse(io.BytesIO(container.encode()))
    r.status_code = 200
    return r



from django.contrib.auth.models import Group, User
from django.test import TestCase
from django.urls import reverse

import hashlib

from scidatacontainer_db.models import DataSet
from . import TESTDIR


class UpdatePermissionTest(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = User.objects.create_user("testuser")

    def _post(self, url, **kwargs):
        self.client.force_login(self.user)
        response = self.client.post(url, **kwargs)
        return response

    def _get(self, url, **kwargs):
        self.client.force_login(self.user)
        response = self.client.get(url, **kwargs)
        return response

    def _create_test_dataset(self):
        filename = TESTDIR + "example.zdc"
        response = self._post(reverse("scidatacontainer_db:ui-fileupload"),
                              data={"uploadfile":
                                    open(filename, "rb")
                                    }
                              )
        self.assertEqual(response.status_code, 201)
        self.id = DataSet.objects.all()[0].id
        self.hash = hashlib.sha256(open(filename, "rb").read()).hexdigest()

        self.perm_url = reverse("scidatacontainer_db:ui-permission_update",
                                args=[str(self.id)])

    def test_view_url_exists_at_desired_location(self):
        self._create_test_dataset()
        response = self._get("/" + str(self.id) + "/permissions/")
        self.assertEqual(response.status_code, 200)

    def test_view_url_accessible_by_name(self):
        self._create_test_dataset()
        response = self._get(self.perm_url)
        self.assertEqual(response.status_code, 200)

    def test_template(self):
        self._create_test_dataset()
        response = self._get(self.perm_url)
        self.assertTemplateUsed(response,
                                "scidatacontainer_db/update_permissions.html")

    def test_create_permission(self):
        self._create_test_dataset()
        testuser2 = User.objects.create_user("testuser2")
        testuser3 = User.objects.create_user("testuser3")

        testgroup1 = Group.objects.create(name="testgroup1")
        testgroup2 = Group.objects.create(name="testgroup2")

        dataset = DataSet.objects.get(id=self.id)
        self.assertEqual(len(dataset.get_read_perm_user_list()), 0)
        self.assertEqual(len(dataset.get_write_perm_user_list()), 0)
        self.assertEqual(len(dataset.get_read_perm_group_list()), 0)
        self.assertEqual(len(dataset.get_write_perm_group_list()), 0)

        response = self._post(self.perm_url,
                              data={"owner": "",
                                    "newuser": testuser2.username,
                                    "newuser_rw-outlined": "ro",
                                    "newgroup": "",
                                    }
                              )
        self.assertEqual(response.status_code, 302)
        self.assertEqual(len(dataset.get_read_perm_user_list()), 1)
        self.assertIn(testuser2, dataset.get_read_perm_user_list())
        self.assertEqual(len(dataset.get_write_perm_user_list()), 0)
        self.assertEqual(len(dataset.get_read_perm_group_list()), 0)
        self.assertEqual(len(dataset.get_write_perm_group_list()), 0)

        response = self._post(self.perm_url,
                              data={"owner": "",
                                    "newuser": testuser3.username,
                                    "newuser_rw-outlined": "rw",
                                    "newgroup": testgroup1.name,
                                    "newgroup_rw-outlined": "ro",
                                    })
        self.assertEqual(response.status_code, 302)
        self.assertEqual(len(dataset.get_read_perm_user_list()), 1)
        self.assertIn(testuser2, dataset.get_read_perm_user_list())
        self.assertEqual(len(dataset.get_write_perm_user_list()), 1)
        self.assertIn(testuser3, dataset.get_write_perm_user_list())
        self.assertEqual(len(dataset.get_read_perm_group_list()), 1)
        self.assertIn(testgroup1, dataset.get_read_perm_group_list())
        self.assertEqual(len(dataset.get_write_perm_group_list()), 0)

        response = self._post(self.perm_url,
                              data={"owner": "",
                                    "newuser": "",
                                    "newgroup": testgroup2.name,
                                    "newgroup_rw-outlined": "rw",
                                    })
        self.assertEqual(response.status_code, 302)
        self.assertEqual(len(dataset.get_read_perm_user_list()), 1)
        self.assertIn(testuser2, dataset.get_read_perm_user_list())
        self.assertEqual(len(dataset.get_write_perm_user_list()), 1)
        self.assertIn(testuser3, dataset.get_write_perm_user_list())
        self.assertEqual(len(dataset.get_read_perm_group_list()), 1)
        self.assertIn(testgroup1, dataset.get_read_perm_group_list())
        self.assertEqual(len(dataset.get_write_perm_group_list()), 1)
        self.assertIn(testgroup2, dataset.get_write_perm_group_list())

        response = self._post(self.perm_url,
                              data={"owner": "",
                                    "newuser": "testuser4",
                                    "newgroup": "",
                                    "newgroup_rw-outlined": "rw",
                                    })

        self.assertEqual(response.status_code, 302)
        self.assertEqual(len(dataset.get_read_perm_user_list()), 1)
        self.assertIn(testuser2, dataset.get_read_perm_user_list())
        self.assertEqual(len(dataset.get_write_perm_user_list()), 1)
        self.assertIn(testuser3, dataset.get_write_perm_user_list())
        self.assertEqual(len(dataset.get_read_perm_group_list()), 1)
        self.assertIn(testgroup1, dataset.get_read_perm_group_list())
        self.assertEqual(len(dataset.get_write_perm_group_list()), 1)
        self.assertIn(testgroup2, dataset.get_write_perm_group_list())

        msgs = [str(m) for m in list(response.wsgi_request._messages)]
        self.assertEqual(len(msgs), 1)
        self.assertIn("User \"testuser4\" does not exist!", msgs)

        response = self._post(self.perm_url,
                              data={"owner": "",
                                    "newuser": "",
                                    "newgroup": "testgroup3",
                                    })

        self.assertEqual(response.status_code, 302)
        self.assertEqual(len(dataset.get_read_perm_user_list()), 1)
        self.assertIn(testuser2, dataset.get_read_perm_user_list())
        self.assertEqual(len(dataset.get_write_perm_user_list()), 1)
        self.assertIn(testuser3, dataset.get_write_perm_user_list())
        self.assertEqual(len(dataset.get_read_perm_group_list()), 1)
        self.assertIn(testgroup1, dataset.get_read_perm_group_list())
        self.assertEqual(len(dataset.get_write_perm_group_list()), 1)
        self.assertIn(testgroup2, dataset.get_write_perm_group_list())

        msgs = [str(m) for m in list(response.wsgi_request._messages)]
        self.assertEqual(len(msgs), 2)
        self.assertIn("Group \"testgroup3\" does not exist!", msgs)

    def test_update_permission(self):
        self.test_create_permission()

        testuser2 = User.objects.get(username="testuser2")
        testuser3 = User.objects.get(username="testuser3")
        testgroup1 = Group.objects.get(name="testgroup1")
        testgroup2 = Group.objects.get(name="testgroup2")
        dataset = DataSet.objects.get(id=self.id)

        response = self._post(self.perm_url,
                              data={"owner": "",
                                    "newuser": "",
                                    "newgroup": "",
                                    "urw_" + str(testuser2.id): "rw"
                                    })
        self.assertEqual(response.status_code, 302)
        self.assertEqual(len(dataset.get_read_perm_user_list()), 0)
        self.assertNotIn(testuser2, dataset.get_read_perm_user_list())
        self.assertEqual(len(dataset.get_write_perm_user_list()), 2)
        self.assertIn(testuser2, dataset.get_write_perm_user_list())
        self.assertIn(testuser3, dataset.get_write_perm_user_list())
        self.assertEqual(len(dataset.get_read_perm_group_list()), 1)
        self.assertIn(testgroup1, dataset.get_read_perm_group_list())
        self.assertEqual(len(dataset.get_write_perm_group_list()), 1)
        self.assertIn(testgroup2, dataset.get_write_perm_group_list())

        response = self._post(self.perm_url,
                              data={"owner": "",
                                    "newuser": "",
                                    "newgroup": "",
                                    "urw_" + str(testuser3.id): "ro",
                                    "grw_" + testgroup1.name: "rw"
                                    })
        self.assertEqual(response.status_code, 302)
        self.assertEqual(len(dataset.get_read_perm_user_list()), 1)
        self.assertNotIn(testuser2, dataset.get_read_perm_user_list())
        self.assertIn(testuser3, dataset.get_read_perm_user_list())
        self.assertEqual(len(dataset.get_write_perm_user_list()), 1)
        self.assertIn(testuser2, dataset.get_write_perm_user_list())
        self.assertNotIn(testuser3, dataset.get_write_perm_user_list())
        self.assertEqual(len(dataset.get_read_perm_group_list()), 0)
        self.assertNotIn(testgroup1, dataset.get_read_perm_group_list())
        self.assertEqual(len(dataset.get_write_perm_group_list()), 2)
        self.assertIn(testgroup1, dataset.get_write_perm_group_list())
        self.assertIn(testgroup2, dataset.get_write_perm_group_list())

        response = self._post(self.perm_url,
                              data={"owner": "",
                                    "newuser": "",
                                    "newgroup": "",
                                    "grw_" + testgroup2.name: "ro"
                                    })
        self.assertEqual(response.status_code, 302)
        self.assertEqual(len(dataset.get_read_perm_user_list()), 1)
        self.assertNotIn(testuser2, dataset.get_read_perm_user_list())
        self.assertIn(testuser3, dataset.get_read_perm_user_list())
        self.assertEqual(len(dataset.get_write_perm_user_list()), 1)
        self.assertIn(testuser2, dataset.get_write_perm_user_list())
        self.assertNotIn(testuser3, dataset.get_write_perm_user_list())
        self.assertEqual(len(dataset.get_read_perm_group_list()), 1)
        self.assertNotIn(testgroup1, dataset.get_read_perm_group_list())
        self.assertIn(testgroup2, dataset.get_read_perm_group_list())
        self.assertEqual(len(dataset.get_write_perm_group_list()), 1)
        self.assertIn(testgroup1, dataset.get_write_perm_group_list())
        self.assertNotIn(testgroup2, dataset.get_write_perm_group_list())

    def test_delete_permission(self):
        self.test_create_permission()

        testuser2 = User.objects.get(username="testuser2")
        testuser3 = User.objects.get(username="testuser3")
        testgroup1 = Group.objects.get(name="testgroup1")
        testgroup2 = Group.objects.get(name="testgroup2")
        dataset = DataSet.objects.get(id=self.id)

        response = self._post(self.perm_url,
                              data={"owner": "",
                                    "newuser": "",
                                    "newgroup": "",
                                    "delete-user": testuser2.id,
                                    })
        self.assertEqual(response.status_code, 302)
        self.assertEqual(len(dataset.get_read_perm_user_list()), 0)
        self.assertNotIn(testuser2, dataset.get_read_perm_user_list())
        self.assertEqual(len(dataset.get_write_perm_user_list()), 1)
        self.assertIn(testuser3, dataset.get_write_perm_user_list())
        self.assertEqual(len(dataset.get_read_perm_group_list()), 1)
        self.assertIn(testgroup1, dataset.get_read_perm_group_list())
        self.assertEqual(len(dataset.get_write_perm_group_list()), 1)
        self.assertIn(testgroup2, dataset.get_write_perm_group_list())

        response = self._post(self.perm_url,
                              data={"owner": "",
                                    "newuser": "",
                                    "newgroup": "",
                                    "delete-group": testgroup1.name,
                                    })
        self.assertEqual(response.status_code, 302)
        self.assertEqual(len(dataset.get_read_perm_user_list()), 0)
        self.assertNotIn(testuser2, dataset.get_read_perm_user_list())
        self.assertEqual(len(dataset.get_write_perm_user_list()), 1)
        self.assertIn(testuser3, dataset.get_write_perm_user_list())
        self.assertEqual(len(dataset.get_read_perm_group_list()), 0)
        self.assertNotIn(testgroup1, dataset.get_read_perm_group_list())
        self.assertEqual(len(dataset.get_write_perm_group_list()), 1)
        self.assertIn(testgroup2, dataset.get_write_perm_group_list())

        response = self._post(self.perm_url,
                              data={"owner": "",
                                    "newuser": testuser2.username,
                                    "newuser_rw-outlined": "ro",
                                    "newgroup": "",
                                    "urw_" + str(testuser3.id): "ro",
                                    "delete-group": testgroup2.name,
                                    })
        self.assertEqual(response.status_code, 302)
        self.assertEqual(len(dataset.get_read_perm_user_list()), 2)
        self.assertIn(testuser2, dataset.get_read_perm_user_list())
        self.assertIn(testuser3, dataset.get_read_perm_user_list())
        self.assertEqual(len(dataset.get_write_perm_user_list()), 0)
        self.assertEqual(len(dataset.get_read_perm_group_list()), 0)
        self.assertEqual(len(dataset.get_write_perm_group_list()), 0)

    def test_change_owner(self):
        self._create_test_dataset()
        testuser2 = User.objects.create_user("testuser2")
        dataset = DataSet.objects.get(id=self.id)
        self.assertEqual(dataset.owner, self.user)

        response = self._post(self.perm_url,
                              data={"owner": testuser2.username,
                                    "newuser": "",
                                    "newgroup": "",
                                    })
        self.assertRedirects(response, reverse("scidatacontainer_db:ui-index"))
        dataset = DataSet.objects.get(id=self.id)
        self.assertEqual(dataset.owner, testuser2)

        dataset.owner = self.user
        dataset.save()
        response = self._post(self.perm_url,
                              data={"owner": "",
                                    "newuser": self.user.username,
                                    "newuser_rw-outlined": "ro",
                                    "newgroup": "",
                                    })

        response = self._post(self.perm_url,
                              data={"owner": testuser2.username,
                                    "newuser": "",
                                    "newgroup": "",
                                    })
        self.assertRedirects(response, reverse("scidatacontainer_db:ui-detail",
                                               args=[str(self.id)]))
        dataset = DataSet.objects.get(id=self.id)
        self.assertEqual(dataset.owner, testuser2)

        response = self._post(self.perm_url,
                              data={"owner": testuser2.username,
                                    "newuser": "",
                                    "newgroup": "",
                                    })
        self.assertEqual(response.status_code, 403)
        dataset = DataSet.objects.get(id=self.id)
        self.assertEqual(dataset.owner, testuser2)

        dataset.owner = self.user
        dataset.save()
        response = self._post(self.perm_url,
                              data={"owner": "testuser3",
                                    "newuser": "",
                                    "newgroup": "",
                                    })
        self.assertRedirects(response, self.perm_url)
        self.assertEqual(dataset.owner, self.user)
        msgs = [str(m) for m in list(response.wsgi_request._messages)]
        self.assertEqual(len(msgs), 1)
        self.assertIn("User \"testuser3\" does not exist!", msgs)



from rest_framework import serializers

from .models import ContainerType, DataSet, DataSetBase, File, Keyword,\
                    Software


class LinkedContainerTypeSerializer(serializers.HyperlinkedModelSerializer):

    url = serializers.HyperlinkedIdentityField(
            view_name="scidatacontainer_db:api:containertype-detail")

    class Meta:
        model = ContainerType
        fields = ["url", "name", "id", "version"]


class LinkedDataSetSerializer(serializers.HyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(
            view_name="scidatacontainer_db:api:dataset-detail")

    class Meta:
        model = DataSetBase
        fields = ["url", "id"]


class LinkedFileSerializer(serializers.HyperlinkedModelSerializer):

    url = serializers.HyperlinkedIdentityField(
            view_name="scidatacontainer_db:api:file-detail")

    class Meta:
        model = File
        fields = ["url", "name", "size"]


class LinkedKeywordSerializer(serializers.HyperlinkedModelSerializer):

    url = serializers.HyperlinkedIdentityField(
            view_name="scidatacontainer_db:api:keyword-detail")

    class Meta:
        model = File
        fields = ["url", "name"]


class LinkedSoftwareSerializer(serializers.HyperlinkedModelSerializer):

    url = serializers.HyperlinkedIdentityField(
            view_name="scidatacontainer_db:api:software-detail")

    class Meta:
        model = Software
        fields = ["url", "name", "version", "id", "id_type"]


class ContainerTypeSerializer(serializers.ModelSerializer):

    instances = LinkedDataSetSerializer(many=True)

    class Meta:
        model = ContainerType
        fields = ["name", "id", "version", "instances", "dbid"]


class DataSetSerializer(serializers.ModelSerializer):
    container_type = LinkedContainerTypeSerializer()
    content = LinkedFileSerializer(many=True, read_only=True)
    keywords = LinkedKeywordSerializer(many=True, read_only=True)
    used_software = LinkedSoftwareSerializer(many=True, read_only=True)
    replaces = LinkedDataSetSerializer(read_only=True)

    class Meta:
        model = DataSet
        fields = ["id", "upload_time", "replaces", "complete",
                  "valid", "size", "created", "storage_time", "static",
                  "container_type", "hash", "used_software", "model_version",
                  "author", "email", "comment", "title", "keywords",
                  "description", "content"]


class FileSerializer(serializers.ModelSerializer):
    included_in = LinkedDataSetSerializer(many=True)

    class Meta:
        model = File
        fields = "__all__"


class KeywordSerializer(serializers.ModelSerializer):

    class Meta:
        model = Keyword
        fields = "__all__"


class SoftwareSerializer(serializers.ModelSerializer):
    used_by = LinkedDataSetSerializer(many=True)

    class Meta:
        model = Software
        fields = "__all__"



from django.urls import reverse
from django.contrib.auth.models import User

import uuid

from rest_framework.test import APITestCase
from knox.models import AuthToken


class ApiAuthenticationTest(APITestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.password = str(uuid.uuid4())
        cls.user = User.objects.create_user("testuser", password=cls.password)
        cls.token = AuthToken.objects.create(cls.user)[1]

    def test_token_authentication(self):
        response = self.client.get(reverse("scidatacontainer_db:api:api-root"))
        self.assertEqual(response.status_code, 401)

        self.client.credentials(HTTP_AUTHORIZATION="Token " + self.token)
        response = self.client.get(reverse("scidatacontainer_db:api:api-root"))
        self.assertEqual(response.status_code, 200)

    def test_password_authentication(self):
        response = self.client.get(reverse("scidatacontainer_db:api:api-root"))
        self.assertEqual(response.status_code, 401)

        self.client.login(username=self.user.username, password=self.password)
        response = self.client.get(reverse("scidatacontainer_db:api:api-root"),
                                   headers={"Authorization": "Token " +
                                                             self.token}
                                   )
        self.assertEqual(response.status_code, 200)



from django.test import TestCase
from django.contrib.auth.models import User, Group
from django.core.files.uploadedfile import InMemoryUploadedFile

import mimetypes

from guardian.shortcuts import assign_perm

from scidatacontainer_db.parsers import parse_container_file
from . import TESTDIR


class DataSetPermissionTest(TestCase):

    def _create_temp_from_file(self, filename):
        file_object = open(filename, "rb")

        def getsize(f):
            f.seek(0)
            f.read()
            s = f.tell()
            f.seek(0)
            return s

        name = filename.split('/')[-1]
        content_type, charset = mimetypes.guess_type(name)
        size = getsize(file_object)

        file = InMemoryUploadedFile(file=file_object, name=name,
                                    field_name=None, content_type=content_type,
                                    charset=charset, size=size)
        return file

    def test_dataset_permission(self):
        testuser1 = User.objects.create_user("Testuser1")
        testuser2 = User.objects.create_user("Testuser2")
        testuser3 = User.objects.create_user("Testuser3")

        testgroup1 = Group.objects.create(name="Testgroup1")
        testgroup2 = Group.objects.create(name="Testgroup2")

        file = self._create_temp_from_file(TESTDIR + "example.zdc")
        obj = parse_container_file(file, testuser1)

        self.assertEqual(len(obj.get_read_perm_user_list()), 0)
        self.assertEqual(len(obj.get_write_perm_user_list()), 0)
        self.assertEqual(len(obj.get_read_perm_group_list()), 0)
        self.assertEqual(len(obj.get_write_perm_group_list()), 0)

        assign_perm("view_dataset", testuser2, obj)
        self.assertIn(testuser2, obj.get_read_perm_user_list())
        self.assertEqual(len(obj.get_read_perm_user_list()), 1)
        self.assertEqual(len(obj.get_write_perm_user_list()), 0)
        self.assertEqual(len(obj.get_read_perm_group_list()), 0)
        self.assertEqual(len(obj.get_write_perm_group_list()), 0)

        assign_perm("change_dataset", testuser3, obj)
        self.assertIn(testuser2, obj.get_read_perm_user_list())
        self.assertEqual(len(obj.get_read_perm_user_list()), 1)
        self.assertEqual(len(obj.get_write_perm_user_list()), 1)
        self.assertIn(testuser3, obj.get_write_perm_user_list())
        self.assertEqual(len(obj.get_read_perm_group_list()), 0)
        self.assertEqual(len(obj.get_write_perm_group_list()), 0)

        assign_perm("view_dataset", testgroup1, obj)
        self.assertIn(testuser2, obj.get_read_perm_user_list())
        self.assertEqual(len(obj.get_read_perm_user_list()), 1)
        self.assertEqual(len(obj.get_write_perm_user_list()), 1)
        self.assertIn(testuser3, obj.get_write_perm_user_list())
        self.assertEqual(len(obj.get_read_perm_group_list()), 1)
        self.assertIn(testgroup1, obj.get_read_perm_group_list())
        self.assertEqual(len(obj.get_write_perm_group_list()), 0)

        assign_perm("change_dataset", testgroup2, obj)
        self.assertEqual(len(obj.get_read_perm_user_list()), 1)
        self.assertIn(testuser2, obj.get_read_perm_user_list())
        self.assertEqual(len(obj.get_write_perm_user_list()), 1)
        self.assertIn(testuser3, obj.get_write_perm_user_list())
        self.assertEqual(len(obj.get_read_perm_group_list()), 1)
        self.assertIn(testgroup1, obj.get_read_perm_group_list())
        self.assertEqual(len(obj.get_write_perm_group_list()), 1)
        self.assertIn(testgroup2, obj.get_write_perm_group_list())



from django.contrib.auth.models import Group, User
from django.urls import reverse

import hashlib
import uuid

from guardian.shortcuts import get_objects_for_user, get_objects_for_group

from scidatacontainer_db.models import DataSet
from . import TESTDIR

from rest_framework.test import APITestCase

VIEW_NAME = "scidatacontainer_db:api:dataset-detail"


class ApiDataSetDetailTest(APITestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = User.objects.create_user("testuser")

    def _post(self, url, **kwargs):
        self.client.force_authenticate(self.user)
        response = self.client.post(url, **kwargs)
        return response

    def _get(self, url, **kwargs):
        self.client.force_authenticate(self.user)
        response = self.client.get(url, **kwargs)
        return response

    def _patch(self, url, **kwargs):
        self.client.force_authenticate(self.user)
        response = self.client.patch(url, **kwargs)
        return response

    def _create_test_dataset(self):
        filename = TESTDIR + "example.zdc"
        response = self._post(reverse("scidatacontainer_db:api:dataset-list"),
                              data={"uploadfile":
                                    open(filename, "rb")
                                    }
                              )
        self.assertEqual(response.status_code, 201)
        self.id = DataSet.objects.all()[0].id
        self.hash = hashlib.sha256(open(filename, "rb").read()).hexdigest()

    def test_view_url_exists_at_desired_location(self):
        self._create_test_dataset()
        response = self._patch("/api/datasets/" + str(self.id) + "/")
        self.assertEqual(response.status_code, 200)

    def test_view_url_accessible_by_name(self):
        self._create_test_dataset()
        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]))
        self.assertEqual(response.status_code, 200)

    def test_http_method(self):
        self._create_test_dataset()
        response = self._get(reverse(VIEW_NAME, args=[str(self.id)]))
        self.assertEqual(response.status_code, 200)

        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]))
        self.assertEqual(response.status_code, 200)

        response = self._post(reverse(VIEW_NAME, args=[str(self.id)]))
        self.assertEqual(response.status_code, 405)

        self.client.force_authenticate(self.user)
        response = self.client.put(reverse(VIEW_NAME, args=[str(self.id)]))
        self.assertEqual(response.status_code, 405)

    def test_permission(self):
        self._create_test_dataset()
        testuser2 = User.objects.create_user("testuser2")

        rand_username = str(uuid.uuid4())
        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]),
                               data={"owner": rand_username})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.reason_phrase,
                         "New owner " + rand_username + " does not exist")

        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]),
                               data={"owner": testuser2.username})
        self.assertEqual(response.status_code, 200)

        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]))
        self.assertEqual(response.status_code, 403)

    def test_dataset_patch_users(self):
        testuser2 = User.objects.create_user("testuser2")
        testuser3 = User.objects.create_user("testuser3")

        response = self._get(reverse(VIEW_NAME, args=[uuid.uuid4()]))
        self.assertEqual(response.status_code, 404)

        self._create_test_dataset()
        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]))
        self.assertEqual(response.status_code, 200)

        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]),
                               data={})
        self.assertEqual(response.status_code, 200)

        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]),
                               data={"author": "Maximiliane Musterfrau"})

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.reason_phrase,
                         "The follwoing fields must not be updated: 'author'.")

        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]),
                               data={"author": "Maximiliane Musterfrau",
                                     "id": "12345"})

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.reason_phrase,
                         "The follwoing fields must not be updated: 'author'" +
                         ", 'id'.")

        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]),
                               data={"readonly_users": testuser2.username})
        self.assertEqual(response.status_code, 200)

        obj = DataSet.objects.get(id=self.id)
        self.assertEqual(list(get_objects_for_user(testuser2,
                                                   "view_dataset",
                                                   DataSet)),
                         [obj])

        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]),
                               data={"readonly_users": "Maxi Musterfrau"})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.reason_phrase,
                         "User Maxi Musterfrau does not exist")

        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]),
                               data={"readonly_users": testuser3.username,
                                     "readwrite_users": testuser2.username})
        self.assertEqual(response.status_code, 200)
        obj = DataSet.objects.get(id=self.id)
        self.assertEqual(list(get_objects_for_user(testuser2,
                                                   "view_dataset",
                                                   DataSet)),
                         [])

        self.assertEqual(list(get_objects_for_user(testuser2,
                                                   "change_dataset",
                                                   DataSet)),
                         [obj])
        self.assertEqual(list(get_objects_for_user(testuser3,
                                                   "view_dataset",
                                                   DataSet)),
                         [obj])
        self.assertEqual(list(get_objects_for_user(testuser3,
                                                   "change_dataset",
                                                   DataSet)),
                         [])

        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]),
                               data={"readonly_users": [None]})
        self.assertEqual(response.status_code, 200)
        obj = DataSet.objects.get(id=self.id)
        self.assertEqual(list(get_objects_for_user(testuser2,
                                                   "view_dataset",
                                                   DataSet)),
                         [])

        self.assertEqual(list(get_objects_for_user(testuser2,
                                                   "change_dataset",
                                                   DataSet)),
                         [obj])

        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]),
                               data={"readwrite_users": [None]})
        self.assertEqual(response.status_code, 200)
        obj = DataSet.objects.get(id=self.id)
        self.assertEqual(list(get_objects_for_user(testuser2,
                                                   "view_dataset",
                                                   DataSet)),
                         [])

        self.assertEqual(list(get_objects_for_user(testuser2,
                                                   "change_dataset",
                                                   DataSet)),
                         [])

    def test_dataset_patch_groups(self):
        testgroup1 = Group.objects.create(name="testgroup1")
        testgroup2 = Group.objects.create(name="testgroup2")

        self._create_test_dataset()

        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]),
                               data={"": "Maximiliane Musterfrau"})

        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]),
                               data={"readonly_groups": testgroup1.name})
        self.assertEqual(response.status_code, 200)
        obj = DataSet.objects.get(id=self.id)
        self.assertEqual(list(get_objects_for_group(testgroup1,
                                                    "view_dataset",
                                                    DataSet)),
                         [obj])

        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]),
                               data={"readonly_groups": testgroup2.name,
                                     "readwrite_groups": testgroup1.name})
        self.assertEqual(response.status_code, 200)
        obj = DataSet.objects.get(id=self.id)
        self.assertEqual(list(get_objects_for_group(testgroup1,
                                                    "view_dataset",
                                                    DataSet)),
                         [])

        self.assertEqual(list(get_objects_for_group(testgroup1,
                                                    "change_dataset",
                                                    DataSet)),
                         [obj])
        self.assertEqual(list(get_objects_for_group(testgroup2,
                                                    "view_dataset",
                                                    DataSet)),
                         [obj])
        self.assertEqual(list(get_objects_for_group(testgroup2,
                                                    "change_dataset",
                                                    DataSet)),
                         [])

        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]),
                               data={"readonly_groups": [None],
                                     "readwrite_groups": [None]})
        self.assertEqual(response.status_code, 200)
        obj = DataSet.objects.get(id=self.id)
        self.assertEqual(list(get_objects_for_group(testgroup1,
                                                    "view_dataset",
                                                    DataSet)), [])

        self.assertEqual(list(get_objects_for_group(testgroup1,
                                                    "change_dataset",
                                                    DataSet)), [])
        self.assertEqual(list(get_objects_for_group(testgroup2,
                                                    "view_dataset",
                                                    DataSet)), [])
        self.assertEqual(list(get_objects_for_group(testgroup2,
                                                    "change_dataset",
                                                    DataSet)), [])

    def test_dataset_patch_valid(self):
        self._create_test_dataset()
        obj = DataSet.objects.get(id=self.id)
        self.assertTrue(obj.valid)

        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]),
                               data={"valid": True})
        self.assertEqual(response.status_code, 200)
        self.assertTrue(obj.valid)

        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]),
                               data={"valid": False})
        self.assertEqual(response.status_code, 200)
        obj = DataSet.objects.get(id=self.id)
        self.assertFalse(obj.valid)

        response = self._patch(reverse(VIEW_NAME, args=[str(self.id)]),
                               data={"valid": True})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.reason_phrase,
                         "It is not possible to change the status of a " +
                         "dataset from invalid to valid.")



from django.test import TestCase
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.contrib.auth.models import User

from datetime import datetime, timezone, timedelta
import mimetypes
import uuid

from scidatacontainer_db.models import ContainerType, DataSet, Keyword,\
                                       Software, DataSetBase
from scidatacontainer_db.parsers import parse_container_file
from scidatacontainer_db.utils import MetaDBError
from . import TESTDIR


class DataSetTest(TestCase):

    def test_dataset_creation(self):
        testuser = User.objects.create_user("Testuser")

        container_type = ContainerType.to_ContainerType({"name": "TestType"})

        now = datetime.now().replace(tzinfo=timezone.utc)
        d = DataSet(complete=True,
                    valid=True,
                    size=20000,
                    created=now,
                    storage_time=now,
                    static=False,
                    owner=testuser,
                    container_type=container_type,
                    )
        d.save()

        self.assertTrue(isinstance(d, DataSet))
        self.assertGreater(d.upload_time, now)
        self.assertEqual(d.complete, True)
        self.assertEqual(d.valid, True)
        self.assertEqual(d.size, 20000)
        self.assertEqual(d.created, now)
        self.assertEqual(d.storage_time, now)
        self.assertEqual(d.static, False)
        self.assertEqual(d.owner, testuser)
        self.assertEqual(d.container_type.name, container_type.name)

    def create_dataset(self):
        testuser = User.objects.create_user("Testuser")

        dc = DataSet()
        dc.owner = testuser

        now = datetime.now().replace(tzinfo=timezone.utc)
        d = {}

        d["complete"] = False
        d["valid"] = False
        d["size"] = 200

        d["created"] = now
        d["storage_time"] = now
        d["static"] = False
        ct = ContainerType(name="TestType2", id="TestID", version="1.0")
        ct.save()
        d["container_type"] = ct
        d["hash"] = "jklafljfkhal"
        s1 = Software(name="TestSoftware", version="1.1")
        s1.save()
        s2 = Software(name="TestSoftware2", version="2.0")
        s2.save()
        d["used_software"] = [s1, s2]
        d["model_version"] = "1.0.0"

        d["author"] = "Max Mustermann"
        d["email"] = "max.mustermann@phoenixd.uni-hannover.de"
        d["comment"] = "TestComment"
        d["title"] = "TestTitle"
        kw1 = Keyword(name="TestKeyword")
        kw1.save()
        kw2 = Keyword(name="TestKeyword2")
        kw2.save()
        d["keywords"] = [kw1, kw2]
        d["description"] = "TestDescription"

        d["content"] = []

        dc.update_attributes(d, testuser)
        dc.save()
        return dc, d

    def test_dataset_update(self):
        dc, d = self.create_dataset()
        for key, value in d.items():
            fieldtype = DataSet._meta.get_field(key).get_internal_type()

            if fieldtype == "ManyToManyField":
                for item in getattr(dc, key).all():
                    self.assertIn(item, value)
            else:
                self.assertEqual(getattr(dc, key), value)

        testuser2 = User.objects.create_user("Testuser2")
        with self.assertRaisesMessage(MetaDBError,
                                      "{'error_code': 403, 'msg': \"You " +
                                      "don't have permission to update this " +
                                      "dataset.\"}"):
            dc.update_attributes(d, testuser2)

        dc.complete = True
        dc.save()
        with self.assertRaisesMessage(MetaDBError,
                                      "{'error_code': 409, 'msg': 'Dataset " +
                                      "is marked complete. No further " +
                                      "changes allowed.'}"):
            dc.update_attributes(d, dc.owner)

        dc.complete = False
        d["static"] = True
        del d["hash"]
        with self.assertRaisesMessage(MetaDBError,
                                      "{'error_code': 400, 'msg': 'A static " +
                                      "dataset requires the hash attribute.'}"
                                      ):
            dc.update_attributes(d, dc.owner)

        d["hash"] = "jklafljfkhal"
        dc.hash = "jklafljfkhal"
        dc.static = True
        dc.save()
        with self.assertRaisesMessage(MetaDBError,
                                      "{'error_code': 301, 'msg': 'A static " +
                                      "file requires a unique hash, but " +
                                      "there is already a file with the " +
                                      "same hash and UUID=" + str(dc.id) +
                                      ".', 'object': <DataSet: " +
                                      "DataSet object (" + str(dc.id) + ")>}"):
            dc.update_attributes(d, dc.owner)

        d["static"] = False
        d["storage_time"] -= timedelta(seconds=1)

        with self.assertRaisesMessage(MetaDBError,
                                      "{'error_code': 400, 'msg': 'Server " +
                                      "version of the dataset is newer than " +
                                      "the file you tried to upload.'}"):
            dc.update_attributes(d, dc.owner)

        d["storage_time"] = d["storage_time"] + timedelta(seconds=2)

        dc2 = DataSet()
        dc2.owner = dc.owner
        d["replaces"] = dc
        dc2.update_attributes(d, dc.owner)
        dc2.save()

        self.assertNotEqual(dc2.replaces, None)
        self.assertEqual(dc2.replaces.id, dc.id)
        self.assertEqual(dc.valid, False)
        self.assertTrue(dc.is_replaced)
        self.assertEqual(dc.replaced_by, dc2)

    def _create_temp_from_file(self, filename):
        file_object = open(filename, "rb")

        def getsize(f):
            f.seek(0)
            f.read()
            s = f.tell()
            f.seek(0)
            return s

        name = filename.split('/')[-1]
        content_type, charset = mimetypes.guess_type(name)
        size = getsize(file_object)

        file = InMemoryUploadedFile(file=file_object, name=name,
                                    field_name=None, content_type=content_type,
                                    charset=charset, size=size)
        return file

    def test_dataset_parse_zip_file(self):
        testuser = User.objects.create_user("Testuser")

        file = self._create_temp_from_file(TESTDIR + "example.zdc")
        obj = parse_container_file(file, testuser)
        obj.save()
        timestamp = datetime(2023, 2, 21,
                             9, 47, 31).replace(tzinfo=timezone.utc)
        self.assertEqual(obj.complete, False)
        self.assertEqual(obj.container_type.name, "myImage")
        self.assertEqual(obj.created, timestamp)
        self.assertEqual(obj.hash, "389c050aadb9ea13c488975c035f06ae3b46c92" +
                                   "b89e69d09a3414958f7521ff9")
        self.assertEqual(obj.model_version, "1.0.0")
        self.assertEqual(obj.storage_time, timestamp)
        self.assertEqual(obj.replaces, None)
        self.assertEqual(obj.static, False)
        self.assertEqual(len(obj.used_software.all()), 2)
        self.assertEqual(obj.id, "a31cb576-494f-40c9-af95-e756271278c3")

        self.assertEqual(obj.author, "Reinhard Caspary")
        self.assertEqual(obj.comment, "")
        self.assertEqual(obj.description, "")
        self.assertEqual(obj.email,
                         "reinhard.caspary@phoenixd.uni-hannover.de")
        self.assertEqual(len(obj.keywords.all()), 2)
        self.assertEqual(obj.title, "This is a sample image dataset")

        # updating dataset
        file = self._create_temp_from_file(TESTDIR + "example_update.zdc")
        obj = parse_container_file(file, testuser)
        obj.save()
        self.assertEqual(obj.container_type.name, "myImage2")
        self.assertEqual(obj.storage_time, timestamp + timedelta(seconds=6))
        lib1 = obj.used_software.get(name="TestLibrary")
        lib2 = obj.used_software.get(name="TestLib2")
        self.assertEqual(lib1.version, "0.2")
        self.assertEqual(lib2.version, "1.2")

        id = obj.id
        obj.delete()
        obj = DataSetBase(id=id)
        obj.save()
        self.assertFalse(isinstance(obj, DataSet))
        self.assertTrue(isinstance(obj, DataSetBase))

        obj = parse_container_file(file, testuser)
        obj.save()
        self.assertTrue(isinstance(obj, DataSet))
        obj.delete()

        # invalid datasets
        file = self._create_temp_from_file(TESTDIR + "example_wo_complete.zdc")

        with self.assertRaisesMessage(MetaDBError,
                                      "{'error_code': 400, 'msg': \""
                                      "Attribute 'complete' required in " +
                                      "content.json.\"}"):
            obj = parse_container_file(file, testuser)

        file = self._create_temp_from_file(TESTDIR + "example_wo_author.zdc")

        with self.assertRaisesMessage(MetaDBError,
                                      "{'error_code': 400, 'msg': \"" +
                                      "Attribute 'author' required in" +
                                      " meta.json.\"}"):
            obj = parse_container_file(file, testuser)

        file = self._create_temp_from_file(TESTDIR + "example_faulty_id.zdc")
        with self.assertRaisesMessage(MetaDBError,
                                      "{'error_code': 400, 'msg': \"" +
                                      "ValidationError: ['“" +
                                      "a31cb576-494f-40c9-af95” is not a " +
                                      "valid UUID.']\"}"):
            obj = parse_container_file(file, testuser)

        file = self._create_temp_from_file(TESTDIR + "example_version.zdc")
        with self.assertRaisesMessage(MetaDBError,
                                      "{'error_code': 400, 'msg': 'You tried" +
                                      " to upload a dataset complying " +
                                      "scidatacontainer model version 0.2.19" +
                                      " but the server requires a minimum " +
                                      "model version of 1.0.0'}"):
            obj = parse_container_file(file, testuser)

    def test_dataset_parse_hdf5_file(self):
        pass

    def test_dataset_parse_pxd_file(self):
        testuser = User.objects.create_user("Testuser")

        file = self._create_temp_from_file(TESTDIR + "test_file.py")
        with self.assertRaisesMessage(MetaDBError,
                                      "{'error_code': 415, 'msg': 'File " +
                                      "format has to be hdf5 or zip!'}"):
            parse_container_file(file, testuser)

    def test_replaced(self):
        dc, d = self.create_dataset()
        d["replaces"] = dc

        dc2 = DataSet(id=uuid.uuid4())
        dc2.owner = dc.owner
        dc2.update_attributes(d, dc.owner)
        dc2.save()

        self.assertTrue(dc.is_replaced)
        self.assertFalse(dc2.is_replaced)
        self.assertEqual(dc.replaces, None)
        self.assertEqual(dc.replaced_by, dc2)
        self.assertEqual(dc2.replaces, dc)
        self.assertEqual(dc2.replaced_by, None)
        self.assertTrue(isinstance(dc.replaced_by, DataSet))
        self.assertTrue(isinstance(dc2.replaces, DataSet))

        dc3 = DataSet(id=uuid.uuid4())
        dc3.owner = dc.owner
        with self.assertRaisesMessage(MetaDBError,
                                      "{'error_code': 409, 'msg': 'Failed to" +
                                      " insert replacement relationship. The" +
                                      " object UUID=" + str(dc.id) + " is " +
                                      "already replaced by " +
                                      "UUID=" + str(dc2.id) + ". You might " +
                                      "want to replace this dataset.', " +
                                      "'object': <DataSet: DataSet object (" +
                                      str(dc2.id) + ")>}"):
            dc3.update_attributes(d, dc.owner)

        dc3 = DataSetBase(id=uuid.uuid4())
        d["replaces"] = dc3
        dc4 = DataSet(id=uuid.uuid4())
        dc4.owner = dc.owner
        dc4.update_attributes(d, dc.owner)

        self.assertTrue(dc3.is_replaced)
        self.assertFalse(dc4.is_replaced)
        self.assertEqual(dc3.replaces, None)
        self.assertEqual(dc3.replaced_by, dc4)
        self.assertEqual(dc4.replaces, dc3)
        self.assertEqual(dc4.replaced_by, None)
        self.assertTrue(isinstance(dc3.replaced_by, DataSet))
        self.assertTrue(isinstance(dc4.replaces, DataSetBase))
        self.assertFalse(isinstance(dc4.replaces, DataSet))

        file = self._create_temp_from_file(TESTDIR + "example_replaces.zdc")
        obj = parse_container_file(file, dc.owner)
        obj.save()
        self.assertTrue(isinstance(obj.replaces, DataSetBase))
        self.assertFalse(isinstance(obj.replaces, DataSet))

        file = self._create_temp_from_file(TESTDIR + "example_faulty.zdc")
        with self.assertRaisesMessage(MetaDBError,
                                      "{'error_code': 400, 'msg': \"Failed " +
                                      "to convert '2023-02-21 UTC' using the" +
                                      " default parser. Make sure it has the" +
                                      " right type.\"}"):
            obj = parse_container_file(file, dc.owner)



from django.contrib.auth.models import User, Group
from django.http import FileResponse
from django.shortcuts import get_object_or_404

from rest_framework import mixins
from rest_framework.decorators import action
from rest_framework.viewsets import ReadOnlyModelViewSet

from guardian.shortcuts import get_objects_for_user, remove_perm, assign_perm

from django_filters.rest_framework import DjangoFilterBackend
from django_filters.rest_framework import FilterSet

from .models import ContainerType, DataSet, File, Keyword, Software
from .parsers import parse_container_file
from .utils import ensure_read_permission, ensure_owner, MetaDBError,\
                   APIResponse as Response
from .test_utils import download_test_dataset, get_test_data
from . import serializers


charfield_filters = ["contains", "icontains", "exact", "iexact",
                     "startswith", "istartswith",
                     "endswith", "iendswith", "regex", "iregex"]


datefield_filter = ["exact", "year__gt", "year__lt", "month__gt",
                    "month__lt", "day__gt", "day__lt"]


class DataSetFilter(FilterSet):
    class Meta:
        model = DataSet
        fields = {
                 "title": charfield_filters,
                 "id": charfield_filters,
                 "upload_time": datefield_filter,
                 "author": charfield_filters,
                 "comment": charfield_filters,
                 "description": charfield_filters,
                 "email": charfield_filters,
                 "created": datefield_filter,
                 "storage_time": datefield_filter,
                 # "container_type": charfield_filters,
                 }


class DataSetViewSet(ReadOnlyModelViewSet, mixins.CreateModelMixin):
    serializer_class = serializers.DataSetSerializer
    filter_backends = [DjangoFilterBackend]
    filterset_class = DataSetFilter
    filterset_fields = ["title", "id"]
    queryset = DataSet.objects.all()

    def get_queryset(self):
        user = self.request.user
        q = get_objects_for_user(user, "view_dataset", DataSet) |\
            user.owner_of.all()
        return q.filter(valid=True)

    def create(self, request):
        if len(request.FILES) > 0:
            try:
                parse_container_file(request.FILES["uploadfile"], request.user)
            except MetaDBError as e:
                obj = e.args[0].get("object", False)
                if obj:
                    s = self.serializer_class(obj,
                                              context={'request': request})
                    return Response(s.data,
                                    status=e.args[0]["error_code"],
                                    reason=e.args[0]["msg"])

                return Response(e.args[0]["msg"],
                                status=e.args[0]["error_code"],
                                reason=e.args[0]["msg"])

            return Response(status=201)
        return Response("", status=400,
                        reason="No data file found in your request!")

    def patch(self, request, pk=None):
        dataset = get_object_or_404(DataSet, id=pk)

        ensure_owner(request.user, dataset)

        accepted_fields = ["readonly_users", "readwrite_users",
                           "readonly_groups", "readwrite_groups",
                           "owner", "valid"]
        rejected = []
        d = {}
        for key, value in request.data.items():
            if key not in accepted_fields:
                rejected.append(key)
            else:
                d[key] = value

        if len(rejected) != 0:
            return Response("", status=400,
                            reason="The follwoing fields must not be " +
                            "updated: '" + "', '".join(rejected) + "'.")

        errors = []

        if "readonly_users" in d:
            for user in dataset.get_read_perm_user_list():
                remove_perm("view_dataset", user, dataset)

            ro_users = d["readonly_users"]
            if not isinstance(ro_users, list):
                ro_users = [ro_users]
            for user in ro_users:
                try:
                    user_obj = User.objects.get(username=user)
                    remove_perm("change_dataset", user_obj, dataset)
                    assign_perm("view_dataset", user_obj, dataset)
                except User.DoesNotExist:
                    errors.append("User " + user + " does not exist")

        if "readwrite_users" in d:
            for user in dataset.get_write_perm_user_list():
                remove_perm("change_dataset", user, dataset)

            rw_users = d["readwrite_users"]
            if not isinstance(rw_users, list):
                rw_users = [rw_users]
            for user in rw_users:
                try:
                    user_obj = User.objects.get(username=user)
                    remove_perm("view_dataset", user_obj, dataset)
                    assign_perm("change_dataset", user_obj, dataset)
                except User.DoesNotExist:
                    errors.append("Error: User " + user + " does not exist")

        if "readonly_groups" in d:
            for group in dataset.get_read_perm_group_list():
                remove_perm("view_dataset", group, dataset)

            ro_groups = d["readonly_groups"]
            if not isinstance(ro_groups, list):
                ro_groups = [ro_groups]
            for group in ro_groups:
                try:
                    group_obj = Group.objects.get(name=group)
                    remove_perm("change_dataset", group_obj, dataset)
                    assign_perm("view_dataset", group_obj, dataset)
                except Group.DoesNotExist:
                    errors.append("Group " + group + " does not exist")

        if "readwrite_groups" in d:
            for group in dataset.get_write_perm_group_list():
                remove_perm("change_dataset", group, dataset)

            rw_groups = d["readwrite_groups"]
            if not isinstance(rw_groups, list):
                rw_groups = [rw_groups]
            for group in rw_groups:
                try:
                    group_obj = Group.objects.get(name=group)
                    remove_perm("view_dataset", group_obj, dataset)
                    assign_perm("change_dataset", group_obj, dataset)
                except Group.DoesNotExist:
                    errors.append("Group " + group + " does not exist")

        if "owner" in d:
            try:
                user_obj = User.objects.get(username=d["owner"])
                dataset.owner = user_obj
                dataset.save()
            except User.DoesNotExist:
                errors.append("New owner " + d["owner"] + " does not exist")

        if "valid" in d:
            if d["valid"] and (not dataset.valid):
                errors.append("It is not possible to change the status of a " +
                              "dataset from invalid to valid.")
                return Response("", 400, reason="\n".join(errors))
            dataset.valid = d["valid"]
            dataset.save()
        return Response("", status=200, reason="\n".join(errors))

    def retrieve(self, request, pk=None, status_code=200, redirect=True):
        try:
            obj = DataSet.objects.get(id=pk)
        except DataSet.DoesNotExist:
            if pk.startswith("00000000-0000-0000-0000-00000000"):
                obj = get_test_data(pk)
            else:
                return Response("", status=404,
                                reason="No DataSet with UUID=" +
                                pk + " found!")
        if not obj.valid:
            return Response({"id": str(obj.id),
                             "owner": obj.owner.username,
                             "invalidation_comment": obj.invalidation_comment},
                            status=204, reason="DataSet was deleted!")

        ensure_read_permission(request.user, obj)
        if redirect and obj.is_replaced:
            pk = obj.replaced_by.id
            return self.retrieve(request, pk=pk, status_code=301,
                                 redirect=True)

        serializer = self.get_serializer(obj)
        return Response(serializer.data, status=status_code)

    @action(methods=["get"], detail=True, url_path="noredirect",
            url_name="detail-noredirect")
    def retrieve_noredirect(self, request, pk=None):
        return self.retrieve(request, pk=pk, status_code=200, redirect=False)

    @action(methods=["get"], detail=True, url_path="download",
            url_name="download")
    def download(self, request, pk=None, status_code=200, redirect=True):
        try:
            obj = DataSet.objects.get(id=pk)
        except DataSet.DoesNotExist:
            if pk.startswith("00000000-0000-0000-0000-00000000"):
                return download_test_dataset(pk)
            else:
                return Response("", status=404,
                                reason="No DataSet with UUID=" +
                                pk + " found!")

        if not obj.valid:
            r = Response("", status=204)
            return Response("", status=204, reason="DataSet was deleted!")

        ensure_read_permission(request.user, obj)
        if redirect and obj.is_replaced:
            pk = obj.replaced_by.id
            return self.download(request, pk=pk, status_code=301)

        r = FileResponse(open(obj.server_path, 'rb'))
        r.status_code = status_code
        return r

    @action(methods=["get"], detail=True, url_path="download/noredirect",
            url_name="download-noredirect")
    def download_noredirect(self, request, pk=None):
        return self.download(request, pk=pk, status_code=200, redirect=False)


class PermissionFilteredReadOnlyViewSet(ReadOnlyModelViewSet):
    def get_queryset(self):
        user = self.request.user
        q = get_objects_for_user(user, "view_dataset", DataSet) |\
            user.owner_of.all()
        ids = q.values_list(self.dataset_fieldname + "__" + self.idstr)
        return self.model.objects.filter(**{self.idstr + "__in": ids})


class ContainerTypeViewSet(PermissionFilteredReadOnlyViewSet):
    serializer_class = serializers.ContainerTypeSerializer
    model = ContainerType
    dataset_fieldname = "container_type"
    idstr = "dbid"


class FileViewSet(PermissionFilteredReadOnlyViewSet):
    serializer_class = serializers.FileSerializer
    model = File
    dataset_fieldname = "content"
    idstr = "id"


class KeywordViewSet(PermissionFilteredReadOnlyViewSet):
    serializer_class = serializers.KeywordSerializer
    model = Keyword
    dataset_fieldname = "keywords"
    idstr = "id"


class SoftwareViewSet(PermissionFilteredReadOnlyViewSet):
    serializer_class = serializers.SoftwareSerializer
    model = Software
    dataset_fieldname = "used_software"
    idstr = "dbid"



from django.test import TestCase
from scidatacontainer_db.models import Software


class SoftwareTest(TestCase):

    def test_software_creation(self):
        s = Software(name="TestLibrary", version="1.0")
        s.save()
        self.assertTrue(isinstance(s, Software))
        self.assertEqual(s.name, "TestLibrary")
        self.assertEqual(s.version, "1.0")
        self.assertEqual(s.id, "")
        self.assertEqual(s.id_type, "")

    def test_software_creation_from_dict(self):
        s = Software.to_Software({
                                  "name": "TestLibrary",
                                  "version": "1.0",
                                  "id": "numpy",
                                  "idType": "github",
                                 })
        s.save()
        self.assertTrue(isinstance(s, Software))
        self.assertEqual(s.name, "TestLibrary")
        self.assertEqual(s.version, "1.0")
        self.assertEqual(s.id, "numpy")
        self.assertEqual(s.id_type, "github")

        s = Software.to_Software({
                                  "name": "TestLibrary",
                                  "version": "1.0",
                                 })
        s.save()
        self.assertTrue(isinstance(s, Software))
        self.assertEqual(s.name, "TestLibrary")
        self.assertEqual(s.version, "1.0")
        self.assertEqual(s.id, "")
        self.assertEqual(s.id_type, "")

        with self.assertRaisesMessage(AssertionError, "usedSoftware requires" +
                                                      " version attribute."):
            s = Software.to_Software({
                                      "name": "TestLibrary",
                                     })
            s.save()

        with self.assertRaisesMessage(AssertionError, "usedSoftware requires" +
                                                      " name attribute."):
            s = Software.to_Software({
                                      "version": "1.0",
                                     })
            s.save()

        with self.assertRaisesMessage(AssertionError,
                                      "usedSoftware requires" +
                                      " idType if id is given"):
            s = Software.to_Software({
                                      "name": "TestLibrary",
                                      "version": "1.0",
                                      "id": "numpy",
                                     })
            s.save()



from django.test import TestCase
from django.urls import reverse
from django.contrib.auth.models import User
from django.conf import settings

import uuid


class AuthenticationTest(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.password = str(uuid.uuid4())
        cls.user = User.objects.create_user("testuser", password=cls.password)

    def test_password_authentication(self):
        response = self.client.get(reverse("scidatacontainer_db:ui-index"))
        index_url = reverse("scidatacontainer_db:ui-index")
        self.assertRedirects(response,
                             settings.LOGIN_URL + "?next=" + index_url
                             )

        response = self.client.post(reverse("scidatacontainer_db:ui-login"),
                                    data={"username": self.user.username,
                                          "password": self.password,
                                          "next": index_url,
                                          }
                                    )
        self.assertRedirects(response, index_url)

        response = self.client.get(index_url)
        self.assertEqual(response.status_code, 200)

        response = self.client.get(reverse("scidatacontainer_db:ui-logout"),
                                   follow=True)
        self.assertRedirects(response,
                             settings.LOGIN_URL + "?next=" + index_url
                             )

        response = self.client.get(reverse("scidatacontainer_db:ui-index"))
        self.assertRedirects(response,
                             settings.LOGIN_URL + "?next=" + index_url
                             )



from django.db import models
from django.contrib.auth.models import User, Group

import uuid

from guardian.models import UserObjectPermissionBase, GroupObjectPermissionBase
from guardian.shortcuts import get_users_with_perms, get_groups_with_perms

from .utils import MetaDBError


class Keyword(models.Model):
    """
    Model to represent a Keyword of a DataSet.
    """
    id = models.UUIDField(primary_key=True,
                          default=uuid.uuid4,
                          editable=False,
                          help_text="UUID primary key")
    name = models.CharField(max_length=256,
                            help_text="String representation of a keyword")

    def __str__(self):
        return self.name


class Software(models.Model):
    """
    Model to represent a Software package used in a dataset.
    """
    dbid = models.UUIDField(primary_key=True,
                            default=uuid.uuid4,
                            editable=False,
                            help_text="UUID primary_key")
    name = models.CharField(max_length=256,
                            help_text="Name of the Software package")
    version = models.CharField(max_length=256,
                               help_text="Version of the software package")
    id = models.CharField(max_length=256,
                          help_text="Identifier of the software package")
    id_type = models.CharField(max_length=256,
                               help_text="Type of the given identifier")

    @classmethod
    def to_Software(cls, pt):
        """
        Convert a dictionary as found in a ZDC container into a
        :model:`scidatacontainer_db.Software` instance.
        """
        assert isinstance(pt, dict)
        assert "name" in pt, "usedSoftware requires name attribute."
        assert "version" in pt, "usedSoftware requires version attribute."

        if "id" in pt:
            assert "idType" in pt, "usedSoftware requires idType if " +\
                                   "id is given"
            id = pt["id"]
            id_type = pt["idType"]
        else:
            id = ""
            id_type = ""

        obj, _ = cls.objects.get_or_create(name=pt["name"],
                                           version=pt["version"],
                                           id=id, id_type=id_type)
        return obj


class File(models.Model):
    """
    Model to represent a File from the content of a dataset.
    If the file is a JSON file, the content will be saved, too.
    """
    id = models.UUIDField(primary_key=True,
                          default=uuid.uuid4,
                          editable=False,
                          help_text="UUID primary key")
    name = models.CharField(max_length=256,
                            help_text="File name inside the ZDC dataset")
    size = models.IntegerField(help_text="File size in bytes")
    content = models.JSONField(null=True,
                               blank=True,
                               help_text="Dictionary containing the content " +
                                         "of a JSON file")


class ContainerType(models.Model):
    """
    Model to represent the container type of a dataset.
    """
    dbid = models.UUIDField(primary_key=True,
                            default=uuid.uuid4,
                            editable=False,
                            help_text="UUID primary key")
    name = models.CharField(max_length=256,
                            help_text="Name of the container type")
    id = models.CharField(max_length=256,
                          null=True,
                          blank=True,
                          help_text="Identifier of the container type")
    version = models.CharField(max_length=256,
                               null=True,
                               blank=True,
                               help_text="Version of the container type")

    def __str__(self):
        if self.version:
            return self.name + ", v" + self.version
        return self.name

    @classmethod
    def to_ContainerType(cls, pt):
        """
        Convert a dictionary as found in a ZDC container into a
        :model:`scidatacontainer_db.ContainerType` instance.
        """
        if isinstance(pt, str):
            obj, _ = cls.objects.get_or_create(name=pt, id=None, version=None)
            return obj
        else:
            assert isinstance(pt, dict), "ContainerType needs to be a " +\
                                         "string (name) or a dictionary."
            assert "name" in pt, "ContainerType requires name attribute."

            if "id" in pt:
                assert "version" in pt, "ContainerType requires version" +\
                                        "attribute if id is given"

            id = pt.get("id", None)
            version = pt.get("version", None)
            obj, _ = cls.objects.get_or_create(name=pt["name"],
                                               id=id,
                                               version=version)
            return obj


class DataSetBase(models.Model):
    """
    Base Model of a :model:`scidatacontainer_db.DataSet`. It is required for
    DataSets that are unknown to the server. For example if an unknown ID is
    found in the :model:`scidatacontainer_db.DataSetBase`.replaces attribute,
    an instance of DataSetBase is created. It can be later on replaced by
    uploading this dataset without breaking the relations imposed by
    replaced_by or replaces.
    """
    id = models.UUIDField(primary_key=True,
                          default=uuid.uuid4,
                          editable=False,
                          help_text="UUID primary key")
    _replaced_by_field = models.OneToOneField("self",
                                              related_name="_replaces_field",
                                              blank=True, null=True,
                                              on_delete=models.PROTECT,
                                              help_text="Other instance that" +
                                                        " replaces this " +
                                                        "instance. Don't use" +
                                                        " this field but the" +
                                                        " getter and setter " +
                                                        "functions"
                                              )

    @property
    def is_replaced(self) -> bool:
        """
        Returna true if the model instance is replaced by another DataSet.
        """
        return isinstance(self._replaced_by_field, DataSetBase)

    @property
    def replaced_by(self):
        """
        Other object that replaces this instance. This might be None.
        """
        if self.is_replaced:
            return self._dataset_class_selector(self._replaced_by_field)
        else:
            return None

    @replaced_by.setter
    def replaced_by(self, value):
        self._replaced_by_field = value

    @property
    def replaces(self):
        """
        Other object that is replaced by this instance. This might by None.
        """
        try:
            return self._dataset_class_selector(self._replaces_field)
        except DataSetBase._replaces_field.RelatedObjectDoesNotExist:
            return None

    @replaces.setter
    def replaces(self, replaced_object):
        if replaced_object.is_replaced:
            raise MetaDBError({"error_code": 409,
                               "msg": "Failed to insert replacement " +
                                      "relationship. The object UUID=" +
                                      str(replaced_object.id) + " is already" +
                                      " replaced by UUID=" +
                                      str(replaced_object.replaced_by.id) +
                                      ". You might want to replace this " +
                                      "dataset.",
                               "object": replaced_object.replaced_by})
        replaced_object.replaced_by = self
        replaced_object.save()


class DataSet(DataSetBase):
    """
    Representation of a ZDC dataset after parsing.
    """
    class Meta:
        permissions = (
                ("view_dataset", "Read only access"),
                ("change_dataset", "Read and write access"),
                )
        default_permissions = ()

    owner = models.ForeignKey(User, on_delete=models.PROTECT,
                              related_name="owner_of",
                              blank=True,
                              help_text="Owner of the dataset")

    # database information
    upload_time = models.DateTimeField(auto_now_add=True,
                                       blank=True,
                                       help_text="Datetime of server uplaod")

    complete = models.BooleanField(help_text="Complete flag to ensure " +
                                             "overwrite protection")
    valid = models.BooleanField(help_text="Valid flag to mark erroneous " +
                                          "datasets")
    invalidation_comment = models.TextField(blank=True,
                                            default="",
                                            help_text="Comment to describe " +
                                                      "why this dataset is " +
                                                      "invalid")
    size = models.IntegerField(help_text="Size of the dataset in bytes")
    server_path = models.CharField(max_length=512,
                                   null=True,
                                   blank=True,
                                   help_text="File path to find the file " +
                                             "on the server")

    # content.json
    created = models.DateTimeField(help_text="Creation timestamp of the " +
                                             "container")
    storage_time = models.DateTimeField(help_text="Timestamp of storage" +
                                                  " of the container")
    static = models.BooleanField(default=False,
                                 help_text="Flag for static containers")
    container_type = models.ForeignKey(ContainerType,
                                       on_delete=models.PROTECT,
                                       related_name="instances",
                                       help_text="ContainerType of this " +
                                                 "dataset")
    hash = models.CharField(max_length=256,
                            null=True,
                            blank=True,
                            help_text="Hash of the container")
    used_software = models.ManyToManyField(Software,
                                           related_name="used_by",
                                           blank=True,
                                           help_text="List of software " +
                                                     "packages used in the " +
                                                     "dataset")
    model_version = models.CharField(max_length=256,
                                     help_text="Version of the data model")

    # meta.json
    author = models.CharField(max_length=256)
    email = models.EmailField()
    organization = models.CharField(max_length=1024, blank=True,
                                    help_text="Affiliation of the author")
    comment = models.TextField(blank=True)
    title = models.CharField(max_length=256)
    keywords = models.ManyToManyField(Keyword,
                                      blank=True,
                                      help_text="List of keywords")
    description = models.TextField(blank=True)
    timestamp = models.DateTimeField(blank=True, null=True)
    doi = models.CharField(max_length=512, blank=True)
    license = models.TextField(blank=True,
                               help_text="Name or text of the license")

    # content files
    content = models.ManyToManyField(File,
                                     related_name="included_in",
                                     help_text="List of files included in " +
                                               "the container")

    def get_read_perm_user_list(self):
        """
        Get a list of users that have read only permission for this object.
        """
        return get_users_with_perms(self, with_group_users=False,
                                    only_with_perms_in=["view_dataset"])

    def get_read_perm_group_list(self):
        """
        Get a list of groups that have read only permission for this object.
        """
        perms = get_groups_with_perms(self, attach_perms=True)
        groupnames = [name for name, p in perms.items() if "view_dataset" in p]
        return Group.objects.filter(name__in=groupnames)

    def get_write_perm_user_list(self):
        """
        Get a list of users that have read and write permission for this
        object.
        """
        return get_users_with_perms(self, with_group_users=False,
                                    only_with_perms_in=["change_dataset"])

    def get_write_perm_group_list(self):
        """
        Get a list of groups that have read and write permission for this
        object.
        """
        perms = get_groups_with_perms(self, attach_perms=True)
        groupnames = [name for name, p in perms.items()
                      if "change_dataset" in p]
        return Group.objects.filter(name__in=groupnames)

    def update_attributes(self, d, user):
        """
        Update a :model:`scidatacontainer_db.DataSet` instance with the
        information in the dictionary d. It first ensures that the user is the
        owner of the DataSet.
        """
        if (not user.has_perm("change_dataset", self)) and\
                (self.owner != user):
            raise MetaDBError(
                {"error_code": 403,
                 "msg": "You don't have permission to update this dataset."})

        if self.complete:
            raise MetaDBError({"error_code": 409,
                               "msg": "Dataset is marked complete. " +
                                      "No further changes allowed."})

        if "static" in d:
            if d["static"]:
                if "hash" not in d:
                    raise MetaDBError({"error_code": 400,
                                       "msg": "A static dataset requires " +
                                              "the hash attribute."})
                sobj = DataSet.objects.filter(hash=d["hash"], static=True)

                if len(sobj) != 0:
                    obj = sobj[0]
                    raise MetaDBError({"error_code": 301,
                                       "msg": "A static file requires a " +
                                              "unique hash, but there is " +
                                              "already a file with the same " +
                                              "hash and UUID=" + str(obj.id) +
                                              ".",
                                       "object": obj})

        if self.storage_time:
            if self.storage_time > d["storage_time"]:
                raise MetaDBError({"error_code": 400,
                                   "msg": "Server version of the dataset is" +
                                          " newer than the file you tried " +
                                          "to upload."})

        self.valid = True
        _keys = []
        for key, value in d.items():
            if key == "replaces":
                continue
            else:
                fieldtype = DataSet._meta.get_field(key).get_internal_type()

            if fieldtype == "ManyToManyField":
                # has to be set after save()
                _keys.append(key)
            else:
                setattr(self, key, value)

        self.save()

        if "replaces" in d:
            self.replaces = d["replaces"]

        for key in _keys:
            value = d[key]
            if value and value != []:
                exec("self." + key + ".set(value)")

        self.save()
        return self


def _dataset_class_selector(self, obj):
    """
    Takes a :model:`scidatacontainer_db.DataSetBase` object and returns the
    corresponding :model:`scidatacontainer_db.DataSet` if it exists. Otherwise
    the :model:`scidatacontainer_db.DataSetBase` instance is returned.
    """
    try:
        return DataSet.objects.get(id=obj.id)
    except DataSet.DoesNotExist:
        return obj


DataSetBase._dataset_class_selector = _dataset_class_selector


"""
Classes to make permission checks faster:
https://django-guardian.readthedocs.io/en/stable/userguide/performance.html
"""


class DataSetUserObjectPermission(UserObjectPermissionBase):
    content_object = models.ForeignKey(DataSet, on_delete=models.CASCADE)


class DataSetGroupObjectPermission(GroupObjectPermissionBase):
    content_object = models.ForeignKey(DataSet, on_delete=models.CASCADE)



from django.contrib.auth.models import User
from django.test import TestCase
from django.urls import reverse

import hashlib

from scidatacontainer_db.models import DataSet
from . import TESTDIR


class DetailTest(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = User.objects.create_user("testuser")

    def _post(self, url, **kwargs):
        self.client.force_login(self.user)
        response = self.client.post(url, **kwargs)
        return response

    def _get(self, url, **kwargs):
        self.client.force_login(self.user)
        response = self.client.get(url, **kwargs)
        return response

    def _create_test_dataset(self):
        filename = TESTDIR + "example.zdc"
        response = self._post(reverse("scidatacontainer_db:ui-fileupload"),
                              data={"uploadfile":
                                    open(filename, "rb")
                                    }
                              )
        self.assertEqual(response.status_code, 201)
        self.id = DataSet.objects.all()[0].id
        self.hash = hashlib.sha256(open(filename, "rb").read()).hexdigest()

    def test_view_url_exists_at_desired_location(self):
        self._create_test_dataset()
        response = self._get("/" + str(self.id) + "/")
        self.assertEqual(response.status_code, 200)

    def test_view_url_accessible_by_name(self):
        self._create_test_dataset()
        response = self._get(reverse("scidatacontainer_db:ui-detail",
                                     args=[str(self.id)]))
        self.assertEqual(response.status_code, 200)

    def test_http_method(self):
        self._create_test_dataset()
        response = self._get(reverse("scidatacontainer_db:ui-detail",
                                     args=[str(self.id)]))
        self.assertEqual(response.status_code, 200)

        response = self._post(reverse("scidatacontainer_db:ui-detail",
                                      args=[str(self.id)]))
        self.assertEqual(response.status_code, 405)

    def test_template(self):
        self._create_test_dataset()
        response = self._get(reverse("scidatacontainer_db:ui-detail",
                                     args=[str(self.id)]))
        self.assertTemplateUsed(response, "scidatacontainer_db/detail.html")

    def test_valid_warning(self):
        self._create_test_dataset()
        obj = DataSet.objects.get(id=self.id)
        obj.valid = False
        obj.save()
        response = self._get(reverse("scidatacontainer_db:ui-detail",
                                     args=[str(self.id)]))

        self.assertContains(response,
                            "WARNING: This dataset is marked invalid.")

        obj.invalidation_comment = "Testreason"
        obj.save()

        response = self._get(reverse("scidatacontainer_db:ui-detail",
                                     args=[str(self.id)]))

        self.assertContains(response,
                            "WARNING: This dataset is marked invalid due to " +
                            "the following reason: Testreason")

    def test_invalidation(self):
        self._create_test_dataset()

        response = self._post(reverse("scidatacontainer_db:ui-delete",
                                      args=[str(self.id)]))
        self.assertEqual(response.status_code, 405)

        response = self._post(reverse("scidatacontainer_db:ui-delete",
                                      args=[str(self.id)]),
                              data={"_method": "patch"})

        self.assertRedirects(response, reverse("scidatacontainer_db:ui-detail",
                                               args=[str(self.id)]))

        response = self._post(reverse("scidatacontainer_db:ui-delete",
                                      args=[str(self.id)]),
                              data={"_method": "patch",
                                    "confirm": True,
                                    "reason": "Testreason"},
                              follow=True)

        self.assertRedirects(response, reverse("scidatacontainer_db:ui-index"))
        self.assertContains(response,
                            "Dataset " + str(self.id) + " was deleted.")

        obj = DataSet.objects.get(id=self.id)
        self.assertFalse(obj.valid)
        self.assertEqual(obj.invalidation_comment, "Testreason")

    def test_replaced(self):
        self._create_test_dataset()
        filename = TESTDIR + "example_replaces.zdc"
        response = self._post(reverse("scidatacontainer_db:ui-fileupload"),
                              data={"uploadfile":
                                    open(filename, "rb")
                                    }
                              )
        self.assertEqual(response.status_code, 201)

        dc1 = DataSet.objects.get(id=self.id)
        dc2 = DataSet.objects.exclude(id=self.id)[0]

        response = self._get(reverse("scidatacontainer_db:ui-detail",
                                     args=[str(self.id)]))
        self.assertEqual(response.status_code, 302)

        response = self._get(reverse("scidatacontainer_db:ui-detail",
                                     args=[str(self.id)]), follow=True)
        self.assertContains(response,
                            "The dataset with UUID=" + str(dc1.id) +
                            " was replaced by the dataset with UUID=" +
                            str(dc2.id) + ". You were " +
                            "automatically redirected.",
                            status_code=200)



from django.test import TestCase
import json
from scidatacontainer_db.models import File


class FileTest(TestCase):

    def test_file_creation(self):
        f = File(name="Testfile", size=200)
        f.save()
        self.assertTrue(isinstance(f, File))
        self.assertEqual(f.name, "Testfile")

        json_str = """
{
  "firstName": "John",
  "lastName": "Smith",
  "isAlive": true,
  "age": 27,
  "address": {
    "streetAddress": "21 2nd Street",
    "city": "New York",
    "state": "NY",
    "postalCode": "10021-3100"
  },
  "phoneNumbers": [
    {
      "type": "home",
      "number": "212 555-1234"
    },
    {
      "type": "office",
      "number": "646 555-4567"
    }
  ],
  "children": [
      "Catherine",
      "Thomas",
      "Trevor"
  ],
  "spouse": null
}"""
        f_json = File(name="test.json", size=200, content=json.loads(json_str))
        f_json.save()
        self.assertTrue(isinstance(f_json, File))
        self.assertEqual(f_json.name, "test.json")
        self.assertEqual(f_json.content["firstName"], "John")
        self.assertEqual(f_json.content["lastName"], "Smith")
        self.assertEqual(f_json.content["isAlive"], True)
        self.assertEqual(f_json.content["age"], 27)
        self.assertEqual(f_json.content["address"]["state"], "NY")
        self.assertEqual(f_json.content["phoneNumbers"][0]["type"], "home")
        self.assertEqual(len(f_json.content["phoneNumbers"]), 2)



from django.urls import include, path

from rest_framework import routers

from . import views
from . import api_views

router = routers.DefaultRouter()
router.register(r"datasets", api_views.DataSetViewSet, basename="dataset")
router.register(r"container-types", api_views.ContainerTypeViewSet,
                basename="containertype")
router.register(r"files", api_views.FileViewSet, basename="file")
router.register(r"keywords", api_views.KeywordViewSet, basename="keyword")
router.register(r"softwares", api_views.SoftwareViewSet, basename="software")

app_name = "scidatacontainer_db"
urlpatterns = [
    path('', views.IndexView.as_view(), name='ui-index'),
    path('login/', views.LoginView.as_view(), name='ui-login'),
    path('logout/', views.LogoutView.as_view(), name='ui-logout'),
    path('<uuid:pk>/permissions/', views.UpdatePermissionsView.as_view(),
         name='ui-permission_update'),
    path('<uuid:pk>/delete/', views.InvalidationView.as_view(),
         name='ui-delete'),
    path('<uuid:pk>/', views.DetailView.as_view(), name='ui-detail'),
    path('upload/', views.UploadFileView.as_view(), name='ui-fileupload'),
    path('download/<uuid:pk>', views.DownloadFileView.as_view(),
         name='ui-filedownload'),
    path('keys/', views.ApiKeyView.as_view(), name='ui-keys'),
    path('api/', include((router.urls, app_name), namespace="api")),
]



from django.contrib.auth.models import User
from django.urls import reverse

from scidatacontainer_db.models import DataSet

from rest_framework.test import APITestCase
from . import TESTDIR

VIEW_NAME = "scidatacontainer_db:api:dataset-list"


class ApiUploadTest(APITestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = User.objects.create_user("testuser")

    def _post(self, url, **kwargs):
        self.client.force_authenticate(self.user)
        response = self.client.post(url, **kwargs)
        return response

    def test_view_url_exists_at_desired_location(self):
        response = self._post("/api/datasets/")
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.reason_phrase,
                         "No data file found in your request!")

    def test_view_url_accessible_by_name(self):
        response = self._post(reverse(VIEW_NAME))
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.reason_phrase,
                         "No data file found in your request!")

    def test_http_method(self):
        response = self._post(reverse(VIEW_NAME))
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.reason_phrase,
                         "No data file found in your request!")

    def test_upload(self):
        testid = "a31cb576-494f-40c9-af95-e756271278c3"
        self.assertEqual(len(DataSet.objects.filter(id=testid)), 0)
        response = self._post(reverse(VIEW_NAME),
                              data={"uploadfile":
                                    open(TESTDIR + "example.zdc", "rb")
                                    }
                              )

        self.assertEqual(response.status_code, 201)
        self.assertEqual(len(DataSet.objects.filter(id=testid)), 1)

        response = self._post(reverse(VIEW_NAME),
                              data={"uploadfile":
                                    open(TESTDIR + "example_wo_author.zdc",
                                         "rb")
                                    }
                              )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.reason_phrase,
                         "Attribute 'author' required in meta.json.")

    def test_update(self):
        testid = "a31cb576-494f-40c9-af95-e756271278c3"
        self.assertEqual(len(DataSet.objects.filter(id=testid)), 0)
        response = self._post(reverse(VIEW_NAME),
                              data={"uploadfile":
                                    open(TESTDIR + "example_update.zdc",
                                         "rb")
                                    }
                              )
        self.assertEqual(response.status_code, 201)
        self.assertEqual(len(DataSet.objects.filter(id=testid)), 1)

        response = self._post(reverse(VIEW_NAME),
                              data={"uploadfile":
                                    open(TESTDIR + "example_update.zdc",
                                         "rb")
                                    }
                              )
        self.assertEqual(response.status_code, 201)
        self.assertEqual(len(DataSet.objects.filter(id=testid)), 1)

    def test_static_conflict(self):
        self.assertEqual(len(DataSet.objects.all()), 0)
        response = self._post(reverse(VIEW_NAME),
                              data={"uploadfile":
                                    open(TESTDIR + "example_static.zdc",
                                         "rb")
                                    }
                              )
        self.assertEqual(len(DataSet.objects.all()), 1)

        response = self._post(reverse(VIEW_NAME),
                              data={"uploadfile":
                                    open(TESTDIR + "example_static.zdc",
                                         "rb")
                                    }
                              )
        self.assertEqual(response.status_code, 301)
        self.assertEqual(response.reason_phrase,
                         "A static file requires a unique hash, but there is" +
                         " already a file with the same hash and UUID=" +
                         "a31cb576-494f-40c9-af95-e756271278c3.")



from django.contrib.auth.models import User
from django.test import TestCase
from django.urls import reverse

import hashlib
import uuid

from scidatacontainer_db.models import DataSet
from . import TESTDIR


class DownloadTest(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = User.objects.create_user("testuser")

    def _post(self, url, **kwargs):
        self.client.force_login(self.user)
        response = self.client.post(url, **kwargs)
        return response

    def _get(self, url, **kwargs):
        self.client.force_login(self.user)
        response = self.client.get(url, **kwargs)
        return response

    def _create_test_dataset(self):
        filename = TESTDIR + "example.zdc"
        response = self._post(reverse("scidatacontainer_db:ui-fileupload"),
                              data={"uploadfile":
                                    open(filename, "rb")
                                    }
                              )
        self.assertEqual(response.status_code, 201)
        self.id = DataSet.objects.all()[0].id
        self.hash = hashlib.sha256(open(filename, "rb").read()).hexdigest()

    def test_view_url_exists_at_desired_location(self):
        self._create_test_dataset()
        response = self._get("/download/" + str(self.id))
        self.assertEqual(response.status_code, 200)

    def test_view_url_accessible_by_name(self):
        self._create_test_dataset()
        response = self._get(reverse("scidatacontainer_db:ui-filedownload",
                                     args=[str(self.id)]))
        self.assertEqual(response.status_code, 200)

    def test_http_method(self):
        self._create_test_dataset()
        response = self._get(reverse("scidatacontainer_db:ui-filedownload",
                                     args=[str(self.id)]))
        self.assertEqual(response.status_code, 200)

        response = self._post(reverse("scidatacontainer_db:ui-filedownload",
                                      args=[str(self.id)]))
        self.assertEqual(response.status_code, 405)

    def test_download(self):
        self._create_test_dataset()

        response = self._get(reverse("scidatacontainer_db:ui-filedownload",
                                     args=[str(self.id)]))
        self.assertEqual(response.status_code, 200)
        file_content = b"".join(response.streaming_content)
        return_hash = hashlib.sha256(file_content).hexdigest()

        self.assertEqual(self.hash, return_hash)

        response = self._get(reverse("scidatacontainer_db:ui-filedownload",
                                     args=[str(uuid.uuid4())]))
        self.assertContains(response,
                            "Not Found",
                            status_code=404)



from django.contrib.auth.models import User
from django.urls import reverse

import hashlib
import uuid

from scidatacontainer_db.models import DataSet, DataSetBase
from . import TESTDIR

from rest_framework.test import APITestCase

VIEW_NAME = "scidatacontainer_db:api:dataset-detail"


class ApiDataSetDetailTest(APITestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = User.objects.create_user("testuser")

    def _post(self, url, **kwargs):
        self.client.force_authenticate(self.user)
        response = self.client.post(url, **kwargs)
        return response

    def _get(self, url, **kwargs):
        self.client.force_authenticate(self.user)
        response = self.client.get(url, **kwargs)
        return response

    def _create_test_dataset(self):
        filename = TESTDIR + "example.zdc"
        response = self._post(reverse("scidatacontainer_db:api:dataset-list"),
                              data={"uploadfile":
                                    open(filename, "rb")
                                    }
                              )
        self.assertEqual(response.status_code, 201)
        self.id = DataSet.objects.all()[0].id
        self.hash = hashlib.sha256(open(filename, "rb").read()).hexdigest()

    def test_view_url_exists_at_desired_location(self):
        self._create_test_dataset()
        response = self._get("/api/datasets/" + str(self.id) + "/")
        self.assertEqual(response.status_code, 200)

    def test_view_url_accessible_by_name(self):
        self._create_test_dataset()
        response = self._get(reverse(VIEW_NAME, args=[str(self.id)]))
        self.assertEqual(response.status_code, 200)

    def test_http_method(self):
        self._create_test_dataset()
        response = self._get(reverse(VIEW_NAME, args=[str(self.id)]))
        self.assertEqual(response.status_code, 200)

        response = self._post(reverse(VIEW_NAME, args=[str(self.id)]))
        self.assertEqual(response.status_code, 405)

        self.client.force_authenticate(self.user)
        response = self.client.put(reverse(VIEW_NAME, args=[str(self.id)]))
        self.assertEqual(response.status_code, 405)

    def test_dataset_detail(self):
        rand_id = str(uuid.uuid4())
        response = self._get(reverse(VIEW_NAME, args=[rand_id]))
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.reason_phrase,
                         "No DataSet with UUID=" + rand_id + " found!")

        self._create_test_dataset()
        response = self._get(reverse(VIEW_NAME, args=[str(self.id)]))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["id"], str(self.id))

    def test_replaced_by(self):
        self._create_test_dataset()

        dataset = DataSet.objects.get(id=self.id)
        self.assertFalse(dataset.is_replaced)
        self.assertEqual(dataset.replaced_by, None)

        filename = TESTDIR + "example_replaces.zdc"
        response = self._post(reverse("scidatacontainer_db:api:dataset-list"),
                              data={"uploadfile":
                                    open(filename, "rb")
                                    }
                              )
        self.assertEqual(response.status_code, 201)
        self.assertEqual(len(DataSet.objects.all()), 2)

        dataset = DataSet.objects.get(id=self.id)
        self.assertTrue(dataset.is_replaced)
        self.assertNotEqual(dataset.replaced_by, None)
        self.assertTrue(isinstance(dataset.replaced_by, DataSetBase))

        response = self._get(reverse(VIEW_NAME, args=[str(self.id)]))
        self.assertEqual(response.status_code, 301)
        self.assertNotEqual(response.data["id"], str(self.id))

        url = reverse("scidatacontainer_db:api:dataset-detail-noredirect",
                      args=[str(self.id)])
        response = self._get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["id"], str(self.id))

    def test_invalid(self):
        self._create_test_dataset()
        dataset = DataSet.objects.get(id=self.id)
        self.assertTrue(dataset.valid)

        dataset.valid = False
        dataset.save()
        response = self._get(reverse(VIEW_NAME, args=[str(self.id)]))

        self.assertEqual(response.status_code, 204)
        self.assertEqual(response.reason_phrase, "DataSet was deleted!")



# Generated by Django 4.1.3 on 2023-06-13 09:39

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('auth', '0012_alter_user_first_name_max_length'),
    ]

    operations = [
        migrations.CreateModel(
            name='ContainerType',
            fields=[
                ('dbid', models.UUIDField(default=uuid.uuid4, editable=False, help_text='UUID primary key', primary_key=True, serialize=False)),
                ('name', models.CharField(help_text='Name of the container type', max_length=256)),
                ('id', models.CharField(blank=True, help_text='Identifier of the container type', max_length=256, null=True)),
                ('version', models.CharField(blank=True, help_text='Version of the container type', max_length=256, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='DataSetBase',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, help_text='UUID primary key', primary_key=True, serialize=False)),
                ('_replaced_by_field', models.OneToOneField(blank=True, help_text="Other instance that replaces this instance. Don't use this field but the getter and setter functions", null=True, on_delete=django.db.models.deletion.PROTECT, related_name='_replaces_field', to='scidatacontainer_db.datasetbase')),
            ],
        ),
        migrations.CreateModel(
            name='File',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, help_text='UUID primary key', primary_key=True, serialize=False)),
                ('name', models.CharField(help_text='File name inside the ZDC dataset', max_length=256)),
                ('size', models.IntegerField(help_text='File size in bytes')),
                ('content', models.JSONField(blank=True, help_text='Dictionary containing the content of a JSON file', null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Keyword',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, help_text='UUID primary key', primary_key=True, serialize=False)),
                ('name', models.CharField(help_text='String representation of a keyword', max_length=256)),
            ],
        ),
        migrations.CreateModel(
            name='Software',
            fields=[
                ('dbid', models.UUIDField(default=uuid.uuid4, editable=False, help_text='UUID primary_key', primary_key=True, serialize=False)),
                ('name', models.CharField(help_text='Name of the Software package', max_length=256)),
                ('version', models.CharField(help_text='Version of the software package', max_length=256)),
                ('id', models.CharField(help_text='Identifier of the software package', max_length=256)),
                ('id_type', models.CharField(help_text='Type of the given identifier', max_length=256)),
            ],
        ),
        migrations.CreateModel(
            name='DataSet',
            fields=[
                ('datasetbase_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='scidatacontainer_db.datasetbase')),
                ('upload_time', models.DateTimeField(auto_now_add=True, help_text='Datetime of server uplaod')),
                ('complete', models.BooleanField(help_text='Complete flag to ensure overwrite protection')),
                ('valid', models.BooleanField(help_text='Valid flag to mark erroneous datasets')),
                ('invalidation_comment', models.TextField(blank=True, default='', help_text='Comment to describe why this dataset is invalid')),
                ('size', models.IntegerField(help_text='Size of the dataset in bytes')),
                ('server_path', models.CharField(blank=True, help_text='File path to find the file on the server', max_length=512, null=True)),
                ('created', models.DateTimeField(help_text='Creation timestamp of the container')),
                ('storage_time', models.DateTimeField(help_text='Timestamp of storage of the container')),
                ('static', models.BooleanField(default=False, help_text='Flag for static containers')),
                ('hash', models.CharField(blank=True, help_text='Hash of the container', max_length=256, null=True)),
                ('model_version', models.CharField(help_text='Version of the data model', max_length=256)),
                ('author', models.CharField(max_length=256)),
                ('email', models.EmailField(max_length=254)),
                ('organization', models.CharField(blank=True, help_text='Affiliation of the author', max_length=1024)),
                ('comment', models.TextField(blank=True)),
                ('title', models.CharField(max_length=256)),
                ('description', models.TextField(blank=True)),
                ('timestamp', models.DateTimeField(blank=True, null=True)),
                ('doi', models.CharField(blank=True, max_length=512)),
                ('license', models.TextField(blank=True, help_text='Name or text of the license')),
                ('container_type', models.ForeignKey(help_text='ContainerType of this dataset', on_delete=django.db.models.deletion.PROTECT, related_name='instances', to='scidatacontainer_db.containertype')),
                ('content', models.ManyToManyField(help_text='List of files included in the container', related_name='included_in', to='scidatacontainer_db.file')),
                ('keywords', models.ManyToManyField(blank=True, help_text='List of keywords', to='scidatacontainer_db.keyword')),
                ('owner', models.ForeignKey(blank=True, help_text='Owner of the dataset', on_delete=django.db.models.deletion.PROTECT, related_name='owner_of', to=settings.AUTH_USER_MODEL)),
                ('used_software', models.ManyToManyField(blank=True, help_text='List of software packages used in the dataset', related_name='used_by', to='scidatacontainer_db.software')),
            ],
            options={
                'permissions': (('view_dataset', 'Read only access'), ('change_dataset', 'Read and write access')),
                'default_permissions': (),
            },
            bases=('scidatacontainer_db.datasetbase',),
        ),
        migrations.CreateModel(
            name='DataSetUserObjectPermission',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('permission', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='auth.permission')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
                ('content_object', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='scidatacontainer_db.dataset')),
            ],
            options={
                'abstract': False,
                'unique_together': {('user', 'permission', 'content_object')},
            },
        ),
        migrations.CreateModel(
            name='DataSetGroupObjectPermission',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='auth.group')),
                ('permission', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='auth.permission')),
                ('content_object', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='scidatacontainer_db.dataset')),
            ],
            options={
                'abstract': False,
                'unique_together': {('group', 'permission', 'content_object')},
            },
        ),
    ]



from django.test import TestCase
from django.contrib.auth.models import User

from datetime import datetime, timezone
import uuid

from scidatacontainer_db.models import ContainerType, DataSet, Keyword,\
                                       Software


class DataSetTest(TestCase):

    def test_version_1_0_0(self):
        testuser = User.objects.create_user("Testuser")
        dc = DataSet()
        dc.owner = testuser

        now = datetime.now().replace(tzinfo=timezone.utc)
        d = {}

        d["complete"] = False
        d["valid"] = False
        d["size"] = 200

        d["created"] = now
        d["storage_time"] = now
        d["static"] = False
        ct = ContainerType(name="TestType2", id="TestID", version="1.0")
        ct.save()
        d["container_type"] = ct
        d["hash"] = "jklafljfkhal"
        s1 = Software(name="TestSoftware", version="1.1")
        s1.save()
        s2 = Software(name="TestSoftware2", version="2.0")
        s2.save()
        d["used_software"] = [s1, s2]
        d["model_version"] = "1.0.0"

        d["author"] = "Max Mustermann"
        d["email"] = "max.mustermann@phoenixd.uni-hannover.de"
        d["comment"] = "TestComment"
        d["title"] = "TestTitle"
        kw1 = Keyword(name="TestKeyword")
        kw1.save()
        kw2 = Keyword(name="TestKeyword2")
        kw2.save()
        d["keywords"] = [kw1, kw2]
        d["description"] = "TestDescription"

        d["content"] = []
        d["doi"] = "testdoi"
        d["license"] = "MIT"
        d["model_version"] = "0.5"

        dc.update_attributes(d, testuser)
        dc.save()

        for key, value in d.items():
            fieldtype = DataSet._meta.get_field(key).get_internal_type()

            if fieldtype == "ManyToManyField":
                for item in getattr(dc, key).all():
                    self.assertIn(item, value)
            else:
                self.assertEqual(getattr(dc, key), value)

        return dc, d



from django.conf import settings
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.db import transaction
from django.db.utils import IntegrityError

from abc import ABC, abstractmethod
import datetime
import json
import os.path
import re
import zipfile
import magic
from typing import List

from packaging import version

from .utils import MetaDBError
from .models import ContainerType, DataSet, DataSetBase, File, Keyword,\
                    Software
from .test_utils import parse_test_data


def _datetime_parser(datetime_str: str) -> datetime.datetime:
    """
    Convert an ISO 8601 timestamp string to a python datetime object.

    :param datetime_str: Timestamp string.

    :return: Timestamp as datetime.datetime object.
    """
    try:
        dt = datetime.datetime.fromisoformat(datetime_str)
        return dt
    except ValueError:
        try:
            dt = datetime.datetime.strptime(datetime_str,
                                            "%Y-%m-%dT%H:%M:%S+%z")
            return dt.replace(tzinfo=datetime.timezone.utc)
        except ValueError:
            dt = datetime.datetime.strptime(datetime_str,
                                            "%Y-%m-%d %H:%M:%S %Z")
            return dt.replace(tzinfo=datetime.timezone.utc)


def _used_software_parser(used_software_list: List[dict]) -> List[Software]:
    """
    Convert a list of dictionaries of Softwares to a list of Software objects.

    :param used_software_list: List of dictionaries of Softwares, as found in
    "content.json"

    :return: List of Software objects.
    """
    return [Software.to_Software(entry) for entry in used_software_list]


def _replaces_parser(replaces: str) -> DataSet:
    """
    Return a DataSet object for a given UUID string.

    :param replaces: UUID str of the replaced DataSet.

    :return: DataSet object of the replaced DataSet.
    """
    try:
        return DataSet.objects.get(id=replaces)
    except DataSet.DoesNotExist:
        return DataSetBase.objects.get_or_create(id=replaces)[0]


def _containerType_parser(container_type: dict) -> ContainerType:
    """
    Convert a dictionary about the ContainerType to a ContainerType object.

    :param container_type: Dictionary of ContainerType information.

    :return: ContainerType object.
    """
    return ContainerType.to_ContainerType(container_type)


def _keyword_parser(keywords: List[str]) -> List[Keyword]:
    """
    Convert a list of keyword strings to a list of Keyword objects.

    :param keywords: List of keyword strings.

    :return: List of Keyword obects.
    """
    return [Keyword.objects.get_or_create(name=entry)[0] for entry in keywords]


_constraints_1_0_0 = {
                    "content": {
                                "uuid": (str, True),
                                "replaces": (_replaces_parser, False),
                                "containerType": (_containerType_parser, True),
                                "created": (_datetime_parser, True),
                                "storageTime": (_datetime_parser, True),
                                "static": (bool, True),
                                "complete": (bool, True),
                                "hash": (str, False),
                                "usedSoftware": (_used_software_parser, False),
                                "modelVersion": (str, True)
                                },
                    "meta": {
                             "author": (str, True),
                             "email": (str, True),
                             "comment": (str, False),
                             "title": (str, True),
                             "keywords": (_keyword_parser, False),
                             "description": (str, False),
                             "timestamp": (str, False),
                             "doi": (str, False),
                             "license": (str, False),
                             }
                    }

_constraints = {
                "1.0.0": _constraints_1_0_0,
                }

MIN_SUPPORTED_VERSION = min([version.parse(k) for k in _constraints.keys()])


class BaseParser(ABC):
    """
    Base class for file format specific parsers. Parsers should inherit
    from this class because it ensures the validation of the meta data.
    """
    def __init__(self):
        self.model_version = None

    @property
    def _constraints(self) -> dict:
        """
        Return a dict containing (parser, required) pairs of the single files.

        :raises scidatacontainer_db.MetaDBError: If the model version is not
        supported.

        :return: constraints dictionary.
        """
        if not self.model_version:
            self._read_model_version()
        v = version.parse(self.model_version)
        if v < MIN_SUPPORTED_VERSION:
            raise MetaDBError({"error_code": 400,
                               "msg": "You tried to upload a dataset " +
                                      "complying scidatacontainer model " +
                                      "version " + self.model_version +
                                      " but the server requires a minimum " +
                                      "model version of " +
                                      str(MIN_SUPPORTED_VERSION)
                               })
        key = (max([k for k in _constraints.keys() if version.parse(k) <= v]))
        return _constraints[key]

    @abstractmethod
    def _read_content_json(self):  # pragma: no cover
        """
        Read the content.json file and store its content to self.content.

        This method is file type specific and needs to be overwritten by every
        inheriting class.
        """
        pass

    @abstractmethod
    def _read_meta_json(self):  # pragma: no cover
        """
        Read the meta.json file and store its content to self.meta.

        This method is file type specific and needs to be overwritten by every
        inheriting class.
        """
        pass

    @abstractmethod
    def _read_filelist(self):  # pragma: no cover
        """
        Create a list of File objects and store it to self.files.

        This method is file type specific and needs to be overwritten by every
        inheriting class.
        """
        pass

    def _read_model_version(self):
        """
        Read the model version from content.json and store it as
        self.model_version.
        """
        if "modelVersion" not in self.content:  # pragma: no cover
            self._read_content_json()
        self.model_version = self.content["modelVersion"]

    def _parse_validate(self, filename: str, in_dict: dict) -> dict:
        """
        Validate the content of a dictionary using the constraints for a
        specified file.

        :param filename: Name of the file to validate. Either "meta" or
        "content".

        :param in_dict: Dictionary read from file.

        :raise scidatacontainer_db.MetaDBError: If a required item is not found
        or an error occured during parsing.

        :returns: Dictionary of validated and parsed values.
        """
        c = self._constraints[filename]
        d = {}
        for key, (t, required) in c.items():
            if required and key not in in_dict:
                raise MetaDBError({"error_code": 400,
                                   "msg": "Attribute '" + key +
                                          "' required in " + filename +
                                          ".json."
                                   })
            if key in in_dict:
                if in_dict[key] and in_dict[key] != []:
                    # convert CamelCase to snake_case (more pythonic).
                    name = re.sub(r'(?<!^)(?=[A-Z])', '_', key).lower()
                    try:
                        # try parsing
                        d[name] = t(in_dict[key])
                    except ValueError:
                        raise MetaDBError({"error_code": 400,
                                           "msg": "Failed to convert '" +
                                                  in_dict[key] + "' using " +
                                                  "the default parser. Make " +
                                                  "sure it has the right type."
                                           })
        return d

    def parse(self, filename: str, user: User) -> DataSet:
        """
        Read the meta data from the file, validate it and store it in the DB.

        :param filename: Filename of the ZDC dataset.
        :param user: User sending the request to validate permissions and to
        set ownership.
        """
        self.filename = filename
        self._read_content_json()
        self._read_meta_json()
        self._read_filelist()
        d = {"size": filename.size, "content": self.files}
        d.update(self._parse_validate("content", self.content))
        d.update(self._parse_validate("meta", self.meta))

        uuid = d["uuid"]

        if uuid.startswith("00000000-0000-0000-0000-00000000"):
            # These UUIDs are reserved for testing.
            return parse_test_data(uuid)

        del d["uuid"]
        if len(DataSetBase.objects.filter(id=uuid)) != 0:
            if len(DataSet.objects.filter(id=uuid)) != 0:
                obj = DataSet.objects.get(id=uuid)
                # existing dataset -> try to update
                return obj.update_attributes(d, user)
            else:
                obj = DataSetBase.objects.get(id=uuid)
                # previously dataset only known by ID
                # -> delete and replace with full dataset
                obj.delete()

        obj = DataSet(id=uuid)
        obj.owner = user
        obj.complete = False
        return obj.update_attributes(d, user)


class ZipContainerParser(BaseParser):
    """
    Parser implementing the file type specific routines for .ZIP based
    containers.
    """
    def _read_content_json(self):
        """
        Read the content.json file inside a .ZIP based container.
        """
        with zipfile.ZipFile(self.filename, 'r') as zfile:
            with zfile.open("content.json") as content_json:
                self.content = json.load(content_json)

    def _read_meta_json(self):
        """
        Read the meta.json file inside a .ZIP based container.
        """
        with zipfile.ZipFile(self.filename, 'r') as zfile:
            with zfile.open("meta.json") as meta_json:
                self.meta = json.load(meta_json)

    def _read_filelist(self):
        """
        Create a list of File objects inside a .ZIP container and store it to
        self.files.
        """
        self.files = []
        with zipfile.ZipFile(self.filename, 'r') as zfile:
            for _filename in zfile.namelist():
                size = zfile.getinfo(_filename).file_size
                name = _filename
                if _filename.endswith(".json"):
                    with zfile.open(_filename) as json_file:
                        data = json.load(json_file)
                        file_obj, _ = File.objects.get_or_create(name=name,
                                                                 size=size,
                                                                 content=data,
                                                                 )
                else:
                    file_obj, _ = File.objects.get_or_create(
                                    name=_filename,
                                    size=size,
                                    )
                file_obj.save()

                self.files.append(file_obj)


class Hdf5ContainerParser(BaseParser):

    def _read_content_json(self):
        raise MetaDBError({"error_code": 501,
                           "msg": "The server does not support to parse hdf5" +
                                  " files yet."
                           }
                          )

    def _read_meta_json(self):
        raise MetaDBError({"error_code": 501,
                           "msg": "The server does not support to parse hdf5" +
                                  " files yet."
                           }
                          )


def parse_container_file(filename, owner):
    """
    Find the file type, read the meta data from the file,
    validate it and store it in the DB.

    :param filename: Filename of the ZDC dataset.
    :param user: User sending the request to validate permissions and to
    set ownership.
    """
    try:
        with transaction.atomic():
            filetype = magic.from_buffer(filename.open("rb").read(2048),
                                         mime=True)
            if filetype == "application/zip":
                parser = ZipContainerParser()
                obj = parser.parse(filename, owner)
                #  obj == None for test uploads
                if not obj:
                    return
                server_path = settings.MEDIA_ROOT + "/" + str(obj.id) + ".zdc"
            elif filetype == "application/x-hdf":
                parser = Hdf5ContainerParser()
                obj = parser.parse(filename, owner)
                server_path = settings.MEDIA_ROOT + "/" + str(obj.id) + ".hdf5"
            else:
                raise MetaDBError({"error_code": 415,
                                   "msg": "File format has to be hdf5 or zip!"
                                   }
                                  )
            server_path = os.path.abspath(server_path)
            with open(server_path, 'wb+') as destination:
                for chunk in (filename).chunks():
                    destination.write(chunk)
            obj.server_path = server_path
            obj.save()
            return obj

    except MetaDBError:
        raise
    except IntegrityError as e:  # pragma: no cover
        raise MetaDBError({"error_code": 400,
                           "msg": "IntegrityError: " + str(e)})
    except ValidationError as e:
        raise MetaDBError({"error_code": 400,
                           "msg": "ValidationError: " + str(e)})
    except Exception as e:  # pragma: no cover
        raise MetaDBError({"error_code": 500,
                           "msg": "Unknown error! Please report to your " +
                           "administrator providing these information:" +
                           "\n\n" + str(e)})



from django.contrib.auth.models import User
from django.test import TestCase
from django.urls import reverse

from scidatacontainer_db.models import DataSet
from . import TESTDIR


class UploadTest(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = User.objects.create_user("testuser")

    def _post(self, url, **kwargs):
        self.client.force_login(self.user)
        response = self.client.post(url, **kwargs)
        return response

    def test_view_url_exists_at_desired_location(self):
        response = self._post("/upload/")
        self.assertContains(response,
                            "No data file found!",
                            count=1,
                            status_code=400)

    def test_view_url_accessible_by_name(self):
        response = self._post(reverse("scidatacontainer_db:ui-fileupload"))
        self.assertContains(response,
                            "No data file found!",
                            count=1,
                            status_code=400)

    def test_http_method(self):
        response = self._post(reverse("scidatacontainer_db:ui-fileupload"))
        self.assertContains(response,
                            "No data file found!",
                            count=1,
                            status_code=400)

        self.client.force_login(self.user)
        response = self.client.get(
                                   reverse("scidatacontainer_db:ui-fileupload")
                                   )
        self.assertEqual(response.status_code, 405)

    def test_upload(self):
        testid = "a31cb576-494f-40c9-af95-e756271278c3"
        self.assertEqual(len(DataSet.objects.filter(id=testid)), 0)
        filename = TESTDIR + "example.zdc"
        response = self._post(reverse("scidatacontainer_db:ui-fileupload"),
                              data={"uploadfile": open(filename, "rb")}
                              )

        self.assertContains(response,
                            "Upload successful!",
                            count=1,
                            status_code=201)
        self.assertEqual(len(DataSet.objects.filter(id=testid)), 1)

        filename = TESTDIR + "example_wo_author.zdc"
        response = self._post(reverse("scidatacontainer_db:ui-fileupload"),
                              data={"uploadfile": open(filename, "rb")}
                              )

        self.assertContains(response,
                            "Attribute 'author' required in meta.json.",
                            count=1,
                            status_code=400)



from django.test import LiveServerTestCase
from django.contrib.auth.models import User

from knox.models import AuthToken

from scidatacontainer.tests.test_upload import _TestUpload
from scidatacontainer.tests.test_download import _TestDownload


class TestClientUpload(LiveServerTestCase, _TestUpload):
    __test__ = True

    def setUp(self):
        self.user = User.objects.create_user("testuser")
        self.api_key = AuthToken.objects.create(self.user)[1]


class TestClientDownload(LiveServerTestCase, _TestDownload):
    __test__ = True

    def setUp(self):
        self.user = User.objects.create_user("testuser")
        self.api_key = AuthToken.objects.create(self.user)[1]



from django.test import TestCase
from scidatacontainer_db.models import Keyword


class KeywordTest(TestCase):

    def create_keyword(self, name="Keyword Name"):
        return Keyword(name=name)

    def test_keyword_creation(self):
        kw = self.create_keyword()
        kw.save()
        self.assertTrue(isinstance(kw, Keyword))
        self.assertEqual(str(kw), "Keyword Name")

        kw2 = self.create_keyword("Testname")
        kw2.save()
        self.assertTrue(isinstance(kw2, Keyword))
        self.assertEqual(str(kw2), "Testname")
        self.assertNotEqual(kw2.id, kw.id)



from django.contrib.auth.models import User
from django.test import TestCase
from django.urls import reverse

import hashlib

from scidatacontainer_db.models import DataSet
from . import TESTDIR


class IndexTest(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = User.objects.create_user("testuser")

    def _post(self, url, **kwargs):
        self.client.force_login(self.user)
        response = self.client.post(url, **kwargs)
        return response

    def _get(self, url, **kwargs):
        self.client.force_login(self.user)
        response = self.client.get(url, **kwargs)
        return response

    def _create_test_dataset(self):
        filename = TESTDIR + "example.zdc"
        response = self._post(reverse("scidatacontainer_db:ui-fileupload"),
                              data={"uploadfile":
                                    open(filename, "rb")
                                    }
                              )
        self.assertEqual(response.status_code, 201)
        self.id = DataSet.objects.all()[0].id
        self.hash = hashlib.sha256(open(filename, "rb").read()).hexdigest()

    def test_view_url_exists_at_desired_location(self):
        self._create_test_dataset()
        response = self._get("/")
        self.assertEqual(response.status_code, 200)

    def test_view_url_accessible_by_name(self):
        self._create_test_dataset()
        response = self._get(reverse("scidatacontainer_db:ui-index"))
        self.assertEqual(response.status_code, 200)

    def test_http_method(self):
        self._create_test_dataset()
        response = self._get(reverse("scidatacontainer_db:ui-index"))
        self.assertEqual(response.status_code, 200)

        response = self._post(reverse("scidatacontainer_db:ui-index"))
        self.assertEqual(response.status_code, 405)

    def test_template(self):
        self._create_test_dataset()
        response = self._get(reverse("scidatacontainer_db:ui-index"))
        self.assertTemplateUsed(response, "scidatacontainer_db/index.html")

    def test_search(self):
        self._create_test_dataset()
        response = self._get(reverse("scidatacontainer_db:ui-index"))
        self.assertContains(response,
                            "<td>myImage</td>",
                            count=1,
                            status_code=200)

        response = self._get(reverse("scidatacontainer_db:ui-index"),
                             data={"search": "myImage"})
        self.assertContains(response,
                            "<td>myImage</td>",
                            count=1,
                            status_code=200)

        response = self._get(reverse("scidatacontainer_db:ui-index"),
                             data={"search": "mymage"})
        self.assertContains(response,
                            "<td>myImage</td>",
                            count=0,
                            status_code=200)



from django.contrib.auth.models import User
from django.urls import reverse

from scidatacontainer_db.urls import urlpatterns

from rest_framework.test import APITestCase


class ApiOverviewTest(APITestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = User.objects.create_user("testuser")

    def _get(self, url, **kwargs):
        self.client.force_authenticate(self.user)
        response = self.client.get(url, **kwargs)
        return response

    def test_view_url_exists_at_desired_location(self):
        response = self._get("/api/")
        self.assertEqual(response.status_code, 200)

    def test_view_url_accessible_by_name(self):
        response = self._get(reverse("scidatacontainer_db:api:api-root"))
        self.assertEqual(response.status_code, 200)

    def test_http_methd(self):
        response = self._get(reverse("scidatacontainer_db:api:api-root"))
        self.assertEqual(response.status_code, 200)

        self.client.force_authenticate(self.user)
        response = self.client.post(
                                    reverse("scidatacontainer_db:api:api-root")
                                    )
        self.assertEqual(response.status_code, 405)

    def test_api_overview(self):
        response = self._get(reverse("scidatacontainer_db:api:api-root"))
        content = response.content.decode("utf-8")
        for url in urlpatterns:
            route = url.pattern._route
            if route.startswith("api"):
                route = route.replace("<drf_format_suffix_json_html:format>",
                                      ".json")
                if not route.endswith(".json"):
                    self.assertIn(route, content)



from django.core.exceptions import PermissionDenied
from django.contrib.auth.models import User

from rest_framework.response import Response


def ensure_read_permission(user: User, dataset):
    """
    Checks if the user has read permission for a given DataSet. Raises a
    PermissionDenied Exception otherwise.

    :param user: User object of requesting user.
    :param dataset: DataSet that the user wants to read.

    :raises django.core.exceptions.PermissionDenied: If the user doesn't have
    read permissions.
    """
    if user == dataset.owner:
        return
    if user.has_perm("view_dataset", dataset):
        return
    if user.has_perm("change_dataset", dataset):
        return
    raise PermissionDenied


def ensure_owner(user: User, dataset):
    """
    Checks if the user is the owner of a given DataSet. Raises a
    PermissionDenied Exception otherwise.

    :param user: User object of requesting user.
    :param dataset: DataSet to check the ownership.

    :raises django.core.exceptions.PermissionDenied: If the user isn't the
    owner.
    """
    if user == dataset.owner:
        return
    raise PermissionDenied


class APIResponse(Response):
    """
    Custom Response derived from rest_framewort.response.Response to set
    the reason phrase on creation of the response.
    """
    def __init__(self, data=None, status=None, template_name=None,
                 headers=None, exception=False, content_type=None,
                 reason=None):

        super().__init__(data=data, status=status, template_name=template_name,
                         headers=headers, exception=False, content_type=None)
        if reason:
            self.reason_phrase = reason


class MetaDBError(Exception):
    """
    Excetption raised by various function and caught by the request handling
    views. The status code and reason_phrase stored in the instances will be
    part of the response.
    """
    pass



from django.contrib import admin
from guardian.admin import GuardedModelAdmin

# Register your models here.
from scidatacontainer_db.models import ContainerType, DataSet, File, Keyword,\
                                       Software


class ContainerTypeAdmin(admin.ModelAdmin):
    pass


class DataSetAdmin(GuardedModelAdmin):
    pass


class FileAdmin(admin.ModelAdmin):
    pass


class KeywordAdmin(admin.ModelAdmin):
    pass


class SoftwareAdmin(admin.ModelAdmin):
    pass


admin.site.register(ContainerType, ContainerTypeAdmin)
admin.site.register(DataSet, DataSetAdmin)
admin.site.register(File, FileAdmin)
admin.site.register(Keyword, KeywordAdmin)
admin.site.register(Software, SoftwareAdmin)



from django.conf import settings
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
import django.contrib.auth.views as authviews
from django.contrib.auth.models import User, Group
from django.db.models import Q
from django.http import HttpResponse, HttpResponseRedirect, FileResponse,\
                        HttpResponseNotAllowed
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.html import format_html
from django.views import generic

from knox.models import AuthToken
from guardian.shortcuts import get_objects_for_user, remove_perm, assign_perm

from .models import DataSet
from .parsers import parse_container_file
from .utils import ensure_read_permission, ensure_owner, MetaDBError


def _http_response(msg, status_code):
    r = HttpResponse(msg)
    r.status_code = status_code
    r.reason_phrase = msg
    return r


class UploadFileView(LoginRequiredMixin, generic.View):
    """
    Upload a file and parse its content.
    """
    def post(self, request):
        if len(request.FILES) > 0:
            try:
                parse_container_file(request.FILES["uploadfile"], request.user)
            except MetaDBError as e:
                return _http_response(e.args[0]["msg"],
                                      e.args[0]["error_code"])
            return _http_response("Upload successful!", 201)
        return _http_response("No data file found!", 400)


class DownloadFileView(LoginRequiredMixin, generic.DetailView):
    """
    Provide a ::model::`scidatacontainer_db.DataSet` as a FileResponse.
    """
    model = DataSet

    def get(self, request, pk):
        dataset = get_object_or_404(DataSet, id=pk)
        ensure_read_permission(request.user, dataset)

        return FileResponse(open(dataset.server_path, 'rb'))


class IndexView(LoginRequiredMixin, generic.ListView):
    """
    Display a list of :model:`scidatacontainer_db.DataSet`


    **Template**
    :template:`scidatacontainer_db/index.html`
    """
    paginate_by = 20
    model = DataSet
    template_name = 'scidatacontainer_db/index.html'

    def get_queryset(self):
        """
        Get the list of :model:`scidatacontainer_db.DataSet` objects to
        display. It only provides :model:`scidatacontainer_db.DataSet` objects
        that the user is allowed to read and filters for the search query.
        """
        search = self.request.GET.get("search", False)
        q = get_objects_for_user(self.request.user,
                                 "view_dataset",
                                 DataSet) |\
            get_objects_for_user(self.request.user,
                                 "change_dataset",
                                 DataSet) |\
            self.request.user.owner_of.all()

        q = q.filter(valid=True, _replaced_by_field=None)
        if search:
            q_filt = q.filter(
                              Q(id__icontains=search) |
                              Q(id__icontains=search) |
                              Q(author__icontains=search) |
                              Q(title__icontains=search) |
                              Q(comment__icontains=search) |
                              Q(description__icontains=search) |
                              Q(email__icontains=search) |
                              Q(container_type__name__icontains=search) |
                              Q(hash__icontains=search)
                             )
        else:
            q_filt = q

        return q_filt.order_by('-upload_time')

    def get_context_data(self, **kwargs):
        """
        Context Data function to enable pagination.
        """
        context = super().get_context_data(**kwargs)
        paginator = context["page_obj"].paginator
        page = self.request.GET.get("page", 1)
        context["page_range"] = paginator.get_elided_page_range(number=page,
                                                                on_each_side=2,
                                                                on_ends=0)
        return context


class DetailView(LoginRequiredMixin, generic.DetailView):
    """
    Display the detailed information of a single
    :model:`scidatacontainer_db.DataSet`.

    **Template**
    :template:`scidatacontainer_db/detail.html`
    """
    model = DataSet
    template_name = 'scidatacontainer_db/detail.html'

    def get(self, request, pk):
        """
        Handle a GET request. If this :model:`scidatacontainer_db:DataSet` is
        marked invalid, a warning message will be shown. If it was replaced by
        a successor :model:`scidatacontainer_db.DataSet`, a warning message
        will be shown and the user is redirected to the successor.
        """
        self.object = self.get_object()
        if not self.object.valid:
            if self.object.invalidation_comment != "":
                msg = "WARNING: This dataset is marked invalid due to the " +\
                      "following reason: " + self.object.invalidation_comment
            else:
                msg = "WARNING: This dataset is marked invalid."
            messages.error(self.request, msg)

        if self.object.replaced_by:
            messages.warning(self.request,
                             "The dataset with UUID=" + str(self.object.id) +
                             " was replaced by the dataset with UUID=" +
                             str(self.object.replaced_by.id) + ". You were " +
                             "automatically redirected.")
            return HttpResponseRedirect(
                    reverse("scidatacontainer_db:ui-detail",
                            args=[self.object.replaced_by.id]))
        return super().get(request, pk)

    def get_object(self):
        """
        Get a :model:`scidatacontainer_db.DataSet` instance and make sure the
        requesting user has read permissions.
        """
        dataset = get_object_or_404(DataSet, id=self.kwargs["pk"])
        ensure_read_permission(self.request.user, dataset)
        return dataset


class InvalidationView(LoginRequiredMixin, generic.detail.SingleObjectMixin,
                       generic.View):
    """
    Handle the request from the invalidation modal.

    It handles PATCH and POST requests. POST request require `_method` set
    to `patch` to enable PATCH requests even thought web browsers don't support
    PATCH requests.
    """
    model = DataSet

    def post(self, request, pk):
        """
        Handle a POST request. This is a forward to the patch method if
        `_method` is `patch`.
        """
        if "_method" in self.request.POST:
            if self.request.POST["_method"] == "patch":
                self.request.PATCH = self.request.POST
                return self.patch(request, pk)
        return HttpResponseNotAllowed(["PATCH"])

    def patch(self, request, pk):
        """
        Handle a PATCH request.

        If the user is the owner of a :model:`scidatacontainer_db.DataSet`,
        store the invalidation reason and unset the `valid` flag. Afterwards,
        redirect to the overview page.
        """
        dataset = get_object_or_404(DataSet, id=pk)
        ensure_owner(request.user, dataset)

        if "confirm" in self.request.PATCH:
            dataset.valid = False
            dataset.invalidation_comment = self.request.PATCH["reason"]
            dataset.save()

            messages.info(self.request, "Dataset " + str(pk) + " was deleted.")
            return HttpResponseRedirect(
                    reverse("scidatacontainer_db:ui-index"))

        # confirm checkbox not checked -> don't do anything
        return HttpResponseRedirect(reverse("scidatacontainer_db:ui-detail",
                                            args=[pk]))


class UpdatePermissionsView(LoginRequiredMixin, generic.UpdateView):
    """
    Display a permission modification page and handle a POST request to change
    them.

    **Context**

    ``users``
        List of all available users.

    ``groups``
        List of all available groups.

    ``r_users``
        List of users with read-only permission.

    ``w_users``
        List of users with read and write permission.

    ``rw_users``
        List of users with read-only or read-and-write permission.

    ``r_groups``
        List of groups with read-only permission.

    ``w_groups``
        List of groups with read and write permission.

    ``rw_groups``
        List of groups with read-only or read-and-write permission.

    **Template**

    :template:`scidatacontainer_db/update_permissions.html`
    """
    model = DataSet
    template_name = "scidatacontainer_db/update_permissions.html"
    fields = []

    def get_context_data(self, **kwargs):
        """
        Prepare context data for the template.
        """
        context = super().get_context_data(**kwargs)
        context["users"] = User.objects.all()
        context["groups"] = Group.objects.all()

        r_users = context["dataset"].get_read_perm_user_list()
        w_users = context["dataset"].get_write_perm_user_list()
        context["r_users"] = r_users
        context["w_users"] = w_users
        context["rw_users"] = (r_users | w_users).order_by("username")

        r_groups = context["dataset"].get_read_perm_group_list()
        w_groups = context["dataset"].get_write_perm_group_list()
        context["r_groups"] = r_groups
        context["w_groups"] = w_groups
        context["rw_groups"] = (r_groups | w_groups).order_by("name")

        return context

    def get_object(self):
        """
        Check permission while finding the
        :model:`scidatacontainer_db.DataSet`.
        """
        dataset = get_object_or_404(DataSet, id=self.kwargs["pk"])
        ensure_owner(self.request.user, dataset)
        return dataset

    def post(self, request, pk):
        """
        Handle POST requests to update the permissions on an object.
        """
        dataset = get_object_or_404(DataSet, id=pk)
        ensure_owner(request.user, dataset)
        for key, val in request.POST.items():
            if key.startswith("urw_"):
                user_obj = User.objects.get(id=key[4:])
                remove_perm("change_dataset", user_obj, dataset)
                remove_perm("view_dataset", user_obj, dataset)
                if val == "ro":
                    assign_perm("view_dataset", user_obj, dataset)
                elif val == "rw":
                    assign_perm("change_dataset", user_obj, dataset)

            if key.startswith("grw_"):
                group_obj = Group.objects.get(name=key[4:])
                remove_perm("change_dataset", group_obj, dataset)
                remove_perm("view_dataset", group_obj, dataset)
                if val == "ro":
                    assign_perm("view_dataset", group_obj, dataset)
                elif val == "rw":
                    assign_perm("change_dataset", group_obj, dataset)

        if request.POST["newuser"] != "":
            try:
                user_obj = User.objects.get(username=request.POST["newuser"])
                remove_perm("change_dataset", user_obj, dataset)
                remove_perm("view_dataset", user_obj, dataset)
                if request.POST["newuser_rw-outlined"] == "ro":
                    assign_perm("view_dataset", user_obj, dataset)
                if request.POST["newuser_rw-outlined"] == "rw":
                    assign_perm("change_dataset", user_obj, dataset)
            except User.DoesNotExist:
                messages.error(request, "User \"" + request.POST["newuser"] +
                                        "\" does not exist!")

        if request.POST["newgroup"] != "":
            try:
                group_obj = Group.objects.get(name=request.POST["newgroup"])
                remove_perm("change_dataset", group_obj, dataset)
                remove_perm("view_dataset", group_obj, dataset)
                if request.POST["newgroup_rw-outlined"] == "ro":
                    assign_perm("view_dataset", group_obj, dataset)
                if request.POST["newgroup_rw-outlined"] == "rw":
                    assign_perm("change_dataset", group_obj, dataset)
            except Group.DoesNotExist:
                messages.error(request, "Group \"" + request.POST["newgroup"] +
                                        "\" does not exist!")

        if "delete-user" in request.POST:
            user_obj = User.objects.get(id=request.POST["delete-user"])
            remove_perm("change_dataset", user_obj, dataset)
            remove_perm("view_dataset", user_obj, dataset)

        if "delete-group" in request.POST:
            group_obj = Group.objects.get(name=request.POST["delete-group"])
            remove_perm("change_dataset", group_obj, dataset)
            remove_perm("view_dataset", group_obj, dataset)

        if request.POST["owner"] != "":
            try:
                user_obj = User.objects.get(username=request.POST["owner"])
                dataset.owner = user_obj
                dataset.save()
            except User.DoesNotExist:
                messages.error(request, "User \"" + request.POST["owner"] +
                                        "\" does not exist!")
                return HttpResponseRedirect(reverse("scidatacontainer_db:" +
                                                    "ui-permission_update",
                                                    args=[pk]))
            if self.request.user.has_perm("view_dataset", dataset):
                return HttpResponseRedirect(
                        reverse("scidatacontainer_db:ui-detail",
                                args=[pk]))
            return HttpResponseRedirect(
                    reverse("scidatacontainer_db:ui-index"))

        return HttpResponseRedirect(
                    reverse("scidatacontainer_db:ui-permission_update",
                            args=[pk]))


class LoginView(authviews.LoginView):
    """
    Display a login page.

    **Template**

    :template:`scidatacontainer_db/login.html`
    """
    template_name = "scidatacontainer_db/login.html"


class LogoutView(authviews.LogoutView):
    """
    Logout a user and redirect to index page.
    """
    next_page = settings.LOGOUT_REDIRECT_URL


class ApiKeyView(LoginRequiredMixin, generic.ListView):
    """
    View to list a users API keys. A post request can be used to generate a
    new API key.

    **Template**

    :template:`scidatacontainer_db/api_keys.html`
    """
    paginate_by = 10
    model = AuthToken
    template_name = "scidatacontainer_db/apikeys.html"

    def get_queryset(self):
        """
        Filter :model:`knox.AuthToken` instances to show only the ones from
        the requesting user.
        """
        q = AuthToken.objects.filter(user=self.request.user)
        return q.order_by("-created")

    def get_context_data(self, **kwargs):
        """
        Context Data function to enable pagination.
        """
        context = super().get_context_data(**kwargs)
        paginator = context["page_obj"].paginator
        page = self.request.GET.get("page", 1)
        context["page_range"] = paginator.get_elided_page_range(number=page,
                                                                on_each_side=2,
                                                                on_ends=0)
        return context

    def _delete_token(self, delete):
        """
        Delete a token if it exists and redirect to the API key overview.
        """
        token = get_object_or_404(AuthToken, digest=delete)
        token.delete()
        messages.success(self.request, "API Key deleted!")
        return HttpResponseRedirect(reverse("scidatacontainer_db:ui-keys"))

    def _create_token(self):
        """
        Create a new API key for the requesting user. The new token is
        shown as a message.
        """
        token = AuthToken.objects.create(self.request.user)[1]
        msg = """API Key created! Please remember it, this is the only time
        it is shown! {0:s}
        <button class="btn btn-link btn-msg-copy"
            id="token_{0:s}"
            data-bs-toggle="tooltip" data-vs-placement="top"
            data-bs-title="copy to clipboard">
        <svg class="bi pe-none me-2" width="24" height="24">
        <use xlink:href="#clipboard"/>
        </svg>
        </button>
        """.format(token)
        messages.success(self.request, format_html(msg))
        return HttpResponseRedirect(reverse("scidatacontainer_db:ui-keys"))

    def post(self, request):
        """
        Handle a POST request for key generation or deletion.
        """
        delete = request.POST.get("delete", False)
        if delete:
            return self._delete_token(delete)

        create = request.POST.get("create", False)
        if create:
            return self._create_token()
